#! /usr/bin/python
from __future__ import with_statement
import os
import shutil
from node import Node
from lock import Lock, READING, WRITING
import vcs
import query

# decorators to mark methods as "read" methods, which are idempotent
# and return values immediately, and "write" methods, which need to be
# buffered until the end of the transaction (so they could be rolled
# back).
_readmethods = {}
def read(m):
    _readmethods[m.__name__] = True
    return m

_writemethods = {}
def write(m):
    _writemethods[m.__name__] = True
    return m

# Exceptions and crap
def revision_name(node):
    return vcs.revision_name(node.revision)

class BackendException(Exception):
    pass

class DoesNotExistError(BackendException):
    def __init__(self, node):
        msg = '%s does not exist in %s' %(str(node), revision_name(node))
        super(DoesNotExistError, self).__init__(msg)
        self.node = node


class AlreadyExistsError(BackendException):
    def __init__(self, node, existing=None):
        msg = '%s already exists in %s' %(str(node), revision_name(existing or node))
        super(AlreadyExistsError, self).__init__(msg)
        self.node = node
        self.existing = existing

class Backend(object):
    # The Backend class is the user-facing interface to the VCS. It
    # handles untracked objects, converting from nodes into filenames,
    # and so on. The Models classes interact with a backend rather
    # than directly with the VCS.
    #
    # The backend also maintains some state, usually abstracted into
    # various other classes which are saved and loaded as needed. I
    # considered factoring this into an array of "hook" objects, but
    # decided that might be overkill.

    def __init__(self, vcs):
        self.vcs = vcs
        self._transaction = self.new_transaction()

    def new_transaction(self):
        return Transaction(self)

    @read
    def best_link(self, node, target):
        # Check a subdirectory of the current file first.
        # See Design point #4 for clarification on this behavior.
        rev = node.revision
        parts = node.filename().split('/')
        while parts:
            new = os.path.join(*(parts+[target]))
            if self.vcs.exists(new, rev):
                return node.replace(path=new)
            parts.pop(-1)

        return None

    @read
    def listdir(self, path, revision=None):
        return self.vcs.listdir(path, revision)

    @read
    def current_revision(self):
        return self.vcs.current_revision()

    @read
    def get_part(self, node, part='index'):
        # FIXME: future improvement could be fetching a "whole object" which
        # could load parts on-demand
        # Does a node have a revision? If not, it should
        if node.untracked:
            return file(node.part(part)).read()
        return self.vcs.get(node.part(part), revid=node.revision)

    @read
    def stream_part(self, node, part='index'):
        if node.untracked:
            return file(node.part(part))
        return self.vcs.stream(node.part(part), revid=node.revision)

    @read
    def diff(self, src, dest, node=None):
        return self.vcs.diff(src, dest, node and node.filename())

    @write
    def save_part(self, node, part, value):
        # FIXME: Does the node have a revision? What does it mean if it does?
        self.vcs.dump(node.part(part), value)
        if not node.untracked:
            self.vcs.add(node.part(part))

    @write
    def commit(self, message, nodelist=[]):
        if isinstance(nodelist, Node): nodelist = [nodelist]
        self.vcs.commit([node.filename() for node in nodelist], message)

    @write
    def remove(self, nodelist):
        if isinstance(nodelist, Node):
            nodelist = [nodelist]

        for node in nodelist:
            if node.untracked:
                # FIXME: Feels dangerous
                shutil.rmtree(node.filename())
            else:
                self.vcs.remove(node.filename())

    @read
    def exists(self, node, untracked=False, revision=None):
        '''Nodes are coerced using node.filename(), which checks if any part of
        the node existed in that version. Since every object is expected to
        have an 'index' part, this should be equivalent to using
        node.part('index').'''
        # FIXME: Should we really allow non-node objects here?  I
        # think yes.. but it sure makes this class more complicated!
        # FIXME: right now I think this is only used on Nodes; maybe
        # we can clean this up.
        if isinstance(node, Node):
            untracked = node.untracked
            revision = node.revision
            node = node.filename()

        if untracked:
            return os.path.exists(node)
        return self.vcs.exists(node, revision)

    @read
    def history(self, nodes):
        '''Nodes in this sequence are coerced using node.filename(), since this
        reflects changes to any part of the node.'''
        if isinstance(nodes, Node): nodes = [nodes]
        return self.vcs.history([node.filename() for node in nodes])

    @read
    def list_parts(self, node):
        '''Get all the parts a node has.'''
        if node.untracked:
            return os.listdir(node.filename())
        return self.vcs.listdir(node.filename(), revid=node.revision)

    # FIXME: Maybe a load_commit which returns {'message': ..., 'timestamp':...}
    @read
    def commit_message(self, rev):
        return self.vcs.commit_message(rev)

    @read
    def commit_timestamp(self, rev):
        return self.vcs.commit_timestamp(rev)

_backend = None
def get_backend():
    return _backend

def make_backend(*args, **kwargs):
    global _backend
    if not _backend:
        _backend = Backend(*args, **kwargs)
    return _backend

def unset_backend():
    global _backend
    _backend = None

class RollbackTransaction(Exception):
    '''An "exception" to signify that the current transaction shouldn't be
    applied.'''
    pass

class CircularButterfuckerError(Exception):
    pass

class Transaction(object):
    '''A class which represents a transaction on the backend.

    A Transaction is essentially a fancy proxy for a Backend. Methods on the
    Transaction check whether the corresponding method on the Backend is a
    "read" method or a "write" method. "Read" methods are idempotent and
    can be called immediately, whereas "write" methods are buffered in case
    the transaction is rolled back (by raising RollbackTransaction).

    Transaction is also responsible for locking intelligently. Right
    now I abdicate that responsibility, but we absolutely need to have
    locking because there are race conditions in adding dependencies
    and stuff. This means we can't have transactions in transactions
    and so forth. The good news is that if you try, it'll lock up
    hard.'''
    def __init__(self, backend):
        self.backend = backend
        self.operations = []
        self._autosaved = []
        self._valid = False

    def __getattr__(self, name):
        if not self._valid:
            raise RuntimeError, "transaction is not currently valid"

        if _writemethods.get(name, False):
            return BufferProxy(self, name)

        elif _readmethods.get(name, False):
            return getattr(self.backend, name)

        raise KeyError, "couldn't find method called %s"%(name)

    def __enter__(self):
        # LOCK current directory exclusive for now
        self.lock = Lock('.')
        self.lock.lock(WRITING)
        self._valid = True
        return self

    def apply(self):
        for o in self._autosaved:
            o.save()
        self._autosaved = []

        for o in self.operations:
            name, args, kwargs = o
            getattr(self.backend, name)(*args, **kwargs)
        self.operations = []

    def rollback(self):
        self._autosaved = []
        self.operations = []

    def valid(self):
        return self._valid

    def query(self):
        return query.TransactionQuery(self)

    def autosave(self, obj):
        '''For autosaving untracked objects only..?'''
        self._autosaved.append(obj)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type == None:  # no exceptions
            self.apply()
            # UNLOCK
            self.lock.unlock()
            pass
            self._valid = False

        else:
            self._valid = False
            self.rollback()
            # UNLOCK
            self.lock.unlock()
            # FIXME: is this the right form for issubclass? Use isinstance?
            if issubclass(exc_type, RollbackTransaction):
                # Not really an exception
                return True

class DummyTransaction(Transaction):
    # FIXME: invert the inheritance here? Completely unrelated class?
    def __init__(self, *args):
        pass

    def exists(self, *args):
        return False

    def __getattr__(self, name):
        raise TypeError, "can't call %r on DummyTransaction"%name

DUMMY_TRANSACTION = DummyTransaction()

class BufferProxy(object):
    '''A proxy for a method which just puts the method call in a buffer.

    Takes two arguments -- a target object, and a method name.
    When called, just appends a tuple representing the function call to the
    target's "operations" buffer.
    '''
    def __init__(self, obj, name):
        self.target = obj
        self.name = name

    def __call__(self, *args, **kwargs):
        self.target.operations.append((self.name, args, kwargs))

