# Lock.py -- read/write locking of directories
## It might be wise to support this API, but make all locks "exclusive"
## for the time being until it turns out that we actually have a huge problem
## with concurrency.
## I'm not even sure what kind of concurrency Lockers'll need, eventually, but
## if it's multithreaded it should be easier to build a lock.
import os
import fcntl
import threading
import thread

READING = fcntl.LOCK_SH
WRITING = fcntl.LOCK_EX

CURRENT_THREAD_WRITER = None

class DeadlockException(Exception): pass

class Lock(object):
    def __init__(self, directory):
        self._filelock = FileLock(directory)
        self._threadlock = ThreadLock()

    def lock(self, type):
        global CURRENT_THREAD_WRITER
        if CURRENT_THREAD_WRITER == thread.get_ident():
            raise DeadlockException, "I'm already in a transaction"
        elif type == WRITING:
            CURRENT_THREAD_WRITER = thread.get_ident()
        self.type = type
        self._filelock.lock(type)
        self._threadlock.lock(type)

    def unlock(self):
        global CURRENT_THREAD_WRITER
        if self.type == WRITING:
            CURRENT_THREAD_WRITER = None
        self._filelock.unlock()
        self._threadlock.unlock()

# lockfile1 gets locked whenever any process wants to grab a lock.
# lockfile2 is the lock itself, and can be shared.
#
# lockfile1 is used to prevent writers from getting starved. A writer
# locks lockfile1 so that other readers can't jump on to an
# already-existing shared lock.
#
# Writers keep the lock on lockfile1 for the duration of their time
# being locked. I guess the hope is that this will allow a bunch of
# readers to grab the lock and share it all at once when they're done,
# but practically speaking it doesn't make a difference.
class FileLock(object):
    def __init__(self, directory):
        self.directory = directory
        self.file1 = file(os.path.join(directory, 'lockfile1'), 'w+')
        self.file2 = file(os.path.join(directory, 'lockfile2'), 'w+')

    def lock(self, type):
        fcntl.lockf(self.file1.fileno(), fcntl.LOCK_EX)
        fcntl.lockf(self.file2.fileno(), type)
        self.type = type
        if type == READING:
            fcntl.lockf(self.file1.fileno(), fcntl.LOCK_UN)

    def unlock(self):
        fcntl.lockf(self.file2.fileno(), fcntl.LOCK_UN)
        if self.type == WRITING:
            fcntl.lockf(self.file1.fileno(), fcntl.LOCK_UN)

MAX_READERS = 32  # arbitrary
semaphore = threading.Semaphore(MAX_READERS)
semlock = threading.Lock()

# Algorithm borrowed from http://doc.trolltech.com/qq/qq11-mutex.html
class ThreadLock(object):
    # FIXME: try to detect recursive transactions and throw exception
    def lock(self, type):
        self.type = type
        if self.type == WRITING:
            semlock.acquire()
            for i in xrange(MAX_READERS):
                semaphore.acquire()
            semlock.release()
        else:
            semaphore.acquire()

    def unlock(self):
        if self.type == WRITING:
            for i in xrange(MAX_READERS):
                semaphore.release()
        else:
            semaphore.release()

class StarveException(Exception):
    pass

