import copy
import re
import types
from django.utils.datastructures import SortedDict
from django.core import validators
from django.core.validators import ValidationError
import lockers.format.rst

_component_counter = 0

def add_to_meta(thing, cls, name):
    '''The code shared to Component.contribute_to_class and
    FieldCollector.contribute_to_class.'''
    cls._meta.local_fields.append(thing)
    # If thing is a class, we could use thing.__name__, but let's be consistent
    thing.name = name

# If you think this looks like a stripped-down version of Django's
# Model stuff, you may be right!
class FieldDoesNotExist(Exception):
    pass

class Options(object):
    '''This class is a place to dump fields for objects of FieldCollector.

    The parents attribute exists to keep pointers to the Options for
    parent classes. It has to get set separately by the metaclass.

    In some sense, this class simulates the rules Python uses to look
    up a field in a class while also providing a few methods to
    collect all of the fields in that class.'''
    def __init__(self):
        self.parents = SortedDict()
        self.local_fields = []
        self.local_operations = []

    def contribute_to_class(self, cls, name):
        cls._meta = self
        self.object_name = cls.__name__

    def _fields(self):
        """
        The getter for self.fields. This returns the list of field objects
        available to this model (including through parent models).

        Callers are not permitted to modify this list, since it's a reference
        to this instance (not a copy).
        """
        ## LOCKERS HACK: This should probably rightly be called "components" or
        ## something.
        try:
            self._field_name_cache
        except AttributeError:
            self._fill_fields_cache()
        return self._field_name_cache
    fields = property(_fields)

    def get_fields_with_model(self):
        """
        Returns a dict of name -> (field, model) pairs for all fields. The "model"
        element is None for fields on the current model. Mostly of use when
        constructing queries so that we know which model a field belongs to.
        """
        try:
            self._field_cache
        except AttributeError:
            self._fill_fields_cache()
        return self._field_cache

    def _fill_fields_cache(self):
        # HACKED UP. Django computes the cache of [(field,
        # model)]. But Lockers objects could conceivably replace each
        # other, so we hash on field name.  Need to look at this more.
        cache = {}
        for parent in self.parents:
            for fname, stuff in parent._meta.get_fields_with_model().items():
                field, model = stuff
                if model:
                    cache[fname] = stuff
                else:
                    cache[fname] = (field, parent)
        for field in self.local_fields:
            cache[field.name] = (field, None)

        self._field_cache = cache
        #self._field_name_cache = {}
        self._field_name_cache = []
        for key in cache:
            field, model = cache[key]
            #self._field_name_cache[key] = field
            self._field_name_cache.append(field)


    def get_field(self, name, many_to_many=True, **kwargs):
        """
        Returns the requested field by name. Raises FieldDoesNotExist on error.
        """
        # HACK: many_to_many??
        to_search = many_to_many and (self.fields # + self.many_to_many
                                      ) or self.fields
        for thing in to_search:
            if thing.name == name:
                return thing
        if 'default' in kwargs: return kwargs['default']

        raise FieldDoesNotExist, '%s has no field named %r' % (self.object_name, name)

    def get_field_by_name(self, name):
        """
        Returns the (field_object, model, direct, m2m), where field_object is
        the Field instance for the given name, model is the model containing
        this field (None for local fields), direct is True if the field exists
        on this model, and m2m is True for many-to-many relations. When
        'direct' is False, 'field_object' is the corresponding RelatedObject
        for this field (since the field doesn't have an instance associated
        with it).

        Uses a cache internally, so after the first access, this is very fast.
        """
        try:
            try:
                return self._name_map[name]
            except AttributeError:
                cache = self.init_name_map()
                return cache[name]
        except KeyError:
            raise FieldDoesNotExist('%s has no field named %r'
                    % (self.object_name, name))

    def has_field(self, name):
        return name in self.get_all_field_names()

    def get_all_field_names(self):
        """
        Returns a list of all field names that are possible for this model
        (including reverse relation names).
        """
        try:
            cache = self._name_map
        except AttributeError:
            cache = self.init_name_map()
        names = cache.keys()
        names.sort()
        return names

    def init_name_map(self):
        """
        Initialises the field name -> field object mapping.
        """
        cache = {}
        # We intentionally handle related m2m objects first so that symmetrical
        # m2m accessor names can be overridden, if necessary.
        #for f, model in self.get_all_related_m2m_objects_with_model():
        #    cache[f.field.related_query_name()] = (f, model, False, True)
        #for f, model in self.get_all_related_objects_with_model():
        #    cache[f.field.related_query_name()] = (f, model, False, False)
        #for f, model in self.get_m2m_with_model():
        #    cache[f.name] = (f, model, True, True)
        fields = self.get_fields_with_model()
        for fname in fields:
            cache[fname] = fields[fname] + (True, False)
        #if self.order_with_respect_to:
        #    cache['_order'] = OrderWrt(), None, True, False
        if True or app_cache_ready():
            self._name_map = cache
        return cache

    ### These are really not implemented yet. Django stuff that may not apply.
    def get_add_permission(self):
        return 'add_%s' % self.object_name.lower()

    def get_change_permission(self):
        return 'change_%s' % self.object_name.lower()

    def get_delete_permission(self):
        return 'delete_%s' % self.object_name.lower()


# These need to exist by the time the first class with FieldCollector metaclass
# gets instantiated. These are dummy versions which get replaced later.
class Part(): pass

class InstantiatablePart(): pass

class ListField():
    @classmethod
    def is_valid(cls, thing): return False

class FieldCollector(type):
    '''Metaclass for LockersObjects and Parts.

    These classes are things for whom class definitions mostly just
    specify properties of the object -- namely, fields. This metaclass
    is assigned the task of translating those specifications into
    actual object classes.

    _meta is the place where all the Model-level meta-information is
    stored. It is an instance of an Options object.'''
    _FieldRoots = {}

    def __new__(mcls, name, bases, attrs):
        # Inspired by Django's ModelBase
        super_new = super(FieldCollector, mcls).__new__
        # FIXME: __module__ hack for ModelFactory()
        module = attrs.pop('__module__', '')
        parents = [base for base in bases if isinstance(base, FieldCollector)]

        fields = {}
        parts = {}
        for obj_name in attrs.keys():
            obj = attrs[obj_name]
            if ListField.is_valid(obj):
                attrs[obj_name] = obj = ListField(obj[0])

            # Note: this will fail on creation of the real Part class
            if isinstance(obj, type) and issubclass(obj, Part) or isinstance(obj, InstantiatablePart):
                parts[obj_name] = obj
            elif isinstance(obj, Component):
                fields[obj_name] = attrs.pop(obj_name)

        attrs['__module__'] = module
        new_class = super_new(mcls, name, bases, attrs)
        if parents == []:
            mcls._FieldRoots[new_class] = True

        new_class.add_to_class('_meta', Options())
        for obj_name, obj in fields.items():
            new_class.add_to_class(obj_name, obj)

        for obj_name, obj in parts.items():
            new_class.add_to_class(obj_name, obj)

        for base in parents:
            new_class._meta.parents[base] = True

        return new_class

    def add_to_class(cls, name, value):
        # This has to handle Part subclasses specially. After all,
        # cls.meth is an unbound method. (It has to be available to
        # instances, in case someone does "index = RawPart()".) But we
        # actually want to call the metaclass method, passing the
        # class as "cls" (self). We do that explicitly here.
        if isinstance(value, type) and hasattr(value.__class__, 'contribute_to_class'):
            value.__class__.contribute_to_class(value, cls, name)

        elif hasattr(value, 'contribute_to_class'):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)

    def contribute_to_class(cls, new_cls, name):
        '''Tell this metaclass (probably a subclass of part) to add
        itself to new_cls.'''
        add_to_meta(cls, new_cls, name)

        setattr(new_cls, name, cls)
        cls._meta.object_name = cls.owner_name = new_cls.__name__

    def __repr__(cls):
        for some_cls in cls._FieldRoots:
            if issubclass(cls, some_cls):
                break
        else:
            print "can't happen"
        return '<%s %s named %s at %x>'%(cls.owner_name, some_cls.__name__, cls.__name__, id(cls))

class Component(object):
    def __init__(self, default=None):
        global _component_counter
        self.n = _component_counter
        _component_counter += 1
        self.default = default

    contribute_to_class = add_to_meta

    def __repr__(self):
        if hasattr(self, 'name'):
            return '<%s object named %s at %x>'%(self.__class__.__name__, self.name, id(self))
        return '<unnamed %s object at %x>'%(self.__class__.__name__, id(self))

def set_new_value_for_field(self, kwargs, fields_loc, field, obj, path):
    key = field.name
    newpath = path+'.'+key
    if key in kwargs:
        # class or whatever
        value = kwargs.pop(key)
        value = field.convert(value, obj, newpath)
        if isinstance(field, (type, Part)):
            setattr(self, key, value)
        else:
            setattr(fields_loc, key, value)
    else:
        value = field.new(obj, newpath)
        if isinstance(field, (type, Part)):
            setattr(self, key, value)
        else:
            setattr(fields_loc, key, value)


def fill(self, kwargs, fields_loc, from_=None, object=None, path=''):
    '''Process the keyword arguments for self constructor.

    This goes through self._meta.fields and sets values for each field
    there.  Part subclasses are constructed; Field instances just get
    populated. (See test_model.py for an example.)

    fields_loc is the place values get placed for fields. For Parts,
    this is just the part itself; for LockersObjects, it's self.index.

    @param object: the top-level object that's being filled in
    @param path: the path to get from object to self
    '''
    if from_ == None:
        from_ = self._meta.fields
    if isinstance(kwargs, Part):
        kwargs = kwargs.__dict__

    for field in from_:
        set_new_value_for_field(self, kwargs, fields_loc, field, object, path)

class Part(Component):
    __metaclass__ = FieldCollector
    def __init__(self, *args, **kwargs):
        # Deprecated -- only called by users
        fill(self, kwargs, self)

        if args or kwargs:
            raise TypeError, 'unused arguments to Part: %s %s'%(args, kwargs)

    @classmethod
    def new(cls, object, path):
        return cls()

    @classmethod
    def convert(cls, value, object, path):
        self = cls()
        fill(self, value, self, object=object, path=path)
        return self

    def __repr__(self):
        return "Part(%r)"%(self.__dict__)

    def __eq__(self, rhs):
        if rhs.__class__ == self.__class__ and self.__dict__ == rhs.__dict__:
            return True
        return False

class TuplePart(tuple, Part): pass

class StrPart(str, Part): pass

class InstantiatablePart(Part):
    '''Virtual class for Parts which are expected to be instantiated instead of subclassed.'''
    def __init__(self, *args, **kwargs):
        super(Part, self).__init__()

    def __repr__(self):
        return '<%s instance called %s>'%(self.__class__.__name__, self.name)

class RawPart(InstantiatablePart):
    '''A part representing a single string, which wants to be saved raw rather than YAMLed.

    Originally whether to save an object raw was orthogonal to what
    type it was, but it turned out to only be useful for "file-like
    objects", so it's a separate part now.
    '''
    def __init__(self, stream=True):
        '''stream: whether to stream it from the VCS or just pull it up as a string.'''
        self.stream = stream

    def new(self, object, path):
        return None

    def convert(self, value, object, path):
        return value

class DictPart(InstantiatablePart):
    def new(self, object, path):
        return dict()

    def convert(self, value, object, path):
        return dict(value)


class StrPart(InstantiatablePart):
    def new(self, object, path):
        return ''

    def convert(self, value, object, path):
        return str(value)

class Field(Component):
    def validate(self, field_data, all_data):
        pass

    def new(self, object, path):
        return self.convert(self.default, object, path)

    def convert(self, value, object, path):
        return value

class ListField(Field):
    def __init__(self, type, **kwargs):
        kwargs.setdefault('default', [])
        super(ListField, self).__init__(**kwargs)
        self.type = type

    def convert(self, value, object, path):
        return [self.type.convert(elem, object, path) for elem in value]

    def validate(self, val, all_data):
        pass

    @classmethod
    def is_valid(cls, thing):
        '''Returns true if the thing could be a disguised ListField.

        The common way to disguise a ListField(type) is as [type].'''
        return isinstance(thing, list) and len(thing) == 1 and isinstance(thing[0], Field)

class TupleField(Field):
    def __init__(self, type=None, **kwargs):
        kwargs.setdefault('default', ())
        super(TupleField, self).__init__(**kwargs)

class StrField(Field):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default', '')
        super(StrField, self).__init__(*args, **kwargs)

    def __setter(self, val):
        assert isinstance(val, types.StringTypes)

    def __getter(self):
        return self.val

class TextField(StrField):
    def convert(self, value, object, path):
        return lockers.format.FormattedString(value, lockers.format.rst.RST)

class DictField(Field):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default', {})
        super(DictField, self).__init__(*args, **kwargs)

    def __setter(self, val):
        assert isinstance(val, dict)

    def __getter(self):
        return self.val

class DiffPart(Part):
    old = Field()
    new = Field()

class DiffField(Field):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default', None)
        super(DiffField, self).__init__(*args, **kwargs)

    def convert(self, value, obj, path):
        if value == None:
            return value
        return DiffPart.convert(value, obj, path)

    def new(self, obj, path):
        return None  # I guess??

class NodeField(Field):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default', None)
        super(NodeField, self).__init__(*args, **kwargs)

    def convert(self, value, obj, path):
        if value == None: return value
        try:
            obj.get_node
        except AttributeError:
            return
        if obj.get_node().revision and value.revision == None:
            return value.replace(revision=obj.get_node().revision)
        return value

class EmailField(StrField):
    def validate(self, field_data, all_data):
        validators.isValidEmail(field_data, all_data)
        super(EmailField, self).validate(field_data, all_data)

class NumField(Field):
    def __setter(self, val):
        assert isinstance(val, int)
        self.val = val

    def __getter(self):
        pass

class BoolField(Field):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default', False)
        super(BoolField, self).__init__(*args, **kwargs)

    def __setter(self, val):
        assert isinstance(val, bool)

class DateTimeField(Field):
    def __setter(self, val):
        assert isinstance(val, datetime.datetime)


class FieldOp(object):
    def contribute_to_class(self, cls, name):
        cls._meta.local_operations.append(self)
        self.name = name

class RemovedField(FieldOp):
    def migrate(self, object):
        delattr(object, self.name)

class RenamedField(FieldOp):
    def __init__(self, oldname):
        self.oldname = oldname

    def migrate(self, object):
        # could be expanded to do a split on . and lookup
        value = getattr(self.oldname)

        setattr(object, self.name, value)
