import syck
from lockers.backend import DoesNotExistError, AlreadyExistsError, DummyTransaction
from lockers.dumper import yaml_dump
from lockers import node
from lockers.models import fields
import lockers
import utils

class PRESERVE(object):
    '''Singleton to indicate "don't overwrite whatever's already there"'''
    # FIXME: Not used anywhere? Perhaps get rid of?
    pass

class ModelHolder(object):
    def __init__(self):
        self.versions = {}
        self.latest = -1

    def add(self, ver_no, cls):
        self.versions[ver_no] = cls
        if ver_no > self.latest:
            self.latest = ver_no

    def get(self, ver_no=None):
        if ver_no == None: ver_no = self.latest
        return self.versions[ver_no]

    def has_key(self, ver_no):
        return self.versions.has_key(ver_no)

    def all_versions(self):
        keys = self.versions.keys()
        keys.sort()
        versions = []
        for key in keys:
            versions.append(self.versions[key])

        return versions

    def migrate(self, o, to=None):
        if to == None:
            to = self.latest

        # Invalidate the old o, so that we don't inadvertently save it?

        # FIXME: Need a better iteration model than this
        startver = o.version[0]
        endver = to[0]

        if startver == endver: return o
        for i in range(startver+1, endver+1):
            cls = self.versions[(i,)+o.version[1:]]
            for field in cls._meta.local_fields:
                fields.set_new_value_for_field(o, {}, o.index, field, o, '')

            for operation in cls._meta.local_operations:
                operation.migrate(o)

            if hasattr(o, 'migrate_from_'+str(o.version)):
                getattr(o, 'migrate_from_'+str(o.version))()

            o.version = (i,)+o.version[1:]

        #o = cls.copy(o)   # update the class
        o.__class__ = cls
        return o

def index_getter(propname):
    def getter(self):
        return getattr(self.index, propname)
    return getter

def index_setter(propname):
    def setter(self, val):
        setattr(self.index, propname, val)
    return setter


class AlterPastError(Exception):
    pass

class FactoryType(object): pass
class ModelBase(fields.FieldCollector):
    '''Metaclass for Models.

    In addition to doing FieldsCollector stuff, it might create
    properties for getting/setting fields (and mapping them to the
    Index part). It also keeps track of every class so that it can
    migrate them effectively.

    Factories are also instances of this class.
    '''

    MODEL_CLASSES = {}

    # This type constructs Models classes. This role has expanded from
    # just parsing the field specifications into handling the case
    # where a Models class inherits from a factory. This produces a
    # Cartesian product of classes. If there are two classes Foo_1 and
    # Foo_2, and I create Bar_1 which inherits from Foo_FAC, then the
    # result is Bar_1 (with Foo_1) and Bar_1 (with Foo_2). Bar_1 is
    # then said to be a multiclass.
    #
    # If Bar_2 inherits from Bar_1, by virtue of its being a multiclass, Bar_2
    # must be created in the same way -- Bar_2 (Foo_1) and Bar_2 (Foo_2).
    #
    # Among other things, this code sticks a bunch of stuff in each
    # new class's _meta field. So far this includes:
    #
    # @param multiclass: marks a class that is a "multiclass" --
    # inherits from a Factory or from another multiclass. Note that
    # factories are multiclasses too, and have this attribute set elsewhere.
    #
    # @type multiclass: list of the class made up by this multiclass
    #
    # We also make use of the "parent class" -- the corresponding
    # version of Foo for a given version of Bar.
    #
    # @param parent_class_names: string names of the parent classes
    #
    # @param parent_classes: the parent classes themselves
    def __new__(mcls, name, bases, attrs):
        class_name, class_version = mcls.parse_version(name)

        if issubclass(mcls, FactoryType):
            class_version = -1

        elif class_version == None:
            raise TypeError, "can't figure out what version class %s is"%(name,)

        parent_multiclass = [b for b in bases if isinstance(b, ModelBase) and hasattr(b._meta, 'multiclass')]

        if not parent_multiclass:
            # For example, factories, or LockersObject classes.
            # Just create a new class and stow it.
            ncls = mcls.new_class(name, class_name, bases, attrs)
            if not isinstance(ncls, FactoryType):
                # This should just be LockersObject classes?
                class_version = (class_version,)
                mcls.register_class(class_name, class_version, ncls)

            ncls._meta.class_version = class_version
            ncls._meta.parent_class_names = []
            ncls._meta.parent_classes = []
            return ncls

        # We got some multiclasses in the parents.
        # We should really do some more error checking here.

        # We have to instantiate a class for each combination of versions of
        # each multiclass.
        # These are the versions of each multiclass.
        versions = [b._meta.multiclass for b in parent_multiclass]
        # This is the maximum version number for each multiclass.
        maximum = [len(vlist) for vlist in versions]
        # This is the current version for each multiclass.
        current = [0 for b in parent_multiclass]

        multiclass = []
        while True:
            new_bases = [versions[i][current[i]] for i in range(len(current))]

            ncls = mcls.new_class(name, class_name, tuple(new_bases), attrs)
            ncls._meta.parent_classes = new_bases
            multiclass.append(ncls)

            # FIXME: inheriting from a factory is different from inheriting
            # from some other multiclass. Inheriting from a factory joins
            # version numbers, inheriting from another multiclass means
            # incrementing the last version.
            # Probably we should throw up on inheriting from two multiclasses,
            # except multiple factories would be OK.

            if len(bases) == 1 and not isinstance(bases[0], FactoryType):
                # Inheriting from one singular non-factory multiclass:
                # Probably adding a version to an existing multiclass
                new_version =  (class_version,) + new_bases[0]._meta.class_version[1:]
                ncls._meta.parent_class_names = new_bases[0]._meta.parent_class_names
                # Parent classes are the same as the superclass's parent classes
                ncls._meta.parent_classes = new_bases[0]._meta.parent_classes
            else:
                # Inheriting from a factory or something.
                # Construct a new version number, based on the version numbers
                # of the parent classes, plus this class's version
                new_version = ()
                new_version += (class_version,)
                for i in range(len(current)):
                    new_version += (versions[i][current[i]]._meta.class_version,)
                ncls._meta.parent_class_names = [f._meta.class_name for f in parent_multiclass]
                # Parent classes are the base classes we used to create ncls
                ncls._meta.parent_classes = new_bases

            ncls._meta.class_version = new_version
            #print 'registering', ncls._meta.class_name, ncls._meta.class_version, ncls._meta.parent_class_names
            mcls.register_class(class_name, new_version, ncls)
            utils.advance_list(current, maximum)
            if sum(current) == 0: break

        # Make sure all versions of this class know they're a multiclass
        for cls in multiclass: cls._meta.multiclass = multiclass
        return ncls

    @classmethod
    def register_class(mcls, class_name, class_version, ncls):
        ModelBase.MODEL_CLASSES.setdefault(class_name, ModelHolder())
        ModelBase.MODEL_CLASSES[class_name].add(class_version, ncls)

    @classmethod
    def parse_version(mcls, name):
        n = None
        if '_' not in name: return (name, None)

        head, tail = name.split('_', 1)
        try:
            n = int(tail)
            return (head, n)
        except:
            # FIXME: isn't MyObject_wacky a valid version name?
            return (head, None)

    @classmethod
    def new_class(mcls, name, class_name, bases, attrs):
        attrs = dict(attrs)
        class_fields = {}
        for key in attrs.keys():
            thing = attrs[key]

            # This is a bit of a hack; I reproduce the logic that determines
            # whether the thing I'm looking at belongs in the Index part.
            if isinstance(thing, fields.Field) or \
                    fields.ListField.is_valid(thing):
                class_fields[key] = attrs.pop(key)

        if not attrs.has_key('index'):
            # Create a new Part (a subclass of fields.Part).
            # FIXME: __module__ hack for ModelFactory('foo')
            if '__module__' in attrs:
                class_fields['__module__'] = attrs['__module__']

            spr = fields.Part
            # FIXME: two ModelBase parents -- throw exception
            for base in bases:
                if isinstance(base, ModelBase):
                    spr = base._meta.get_field('index')


            indexcls = fields.FieldCollector('index', (spr,), class_fields)
            attrs['index'] = indexcls

        elif class_fields:
            raise TypeError, 'not implemented: fields + index part'

        # FIXME: Better way to tell if this is a commit class.
        changeset = False
        for base in bases:
            if name == 'Changeset_1' or issubclass(base, Changeset_1):
                changeset = True
                break

        if attrs.has_key('index') and not changeset:
            # Don't create properties that modify index for commits.
            # They don't have Parts anyhow, so you're just modifying one
            # global index object.
            index = attrs['index']
            for part in index._meta.fields:
                part_name = part.name
                attrs[part_name] = property(index_getter(part_name),
                                            index_setter(part_name))
                # FIXME: deleter?

        if attrs.has_key('lockersmeta'):
            for base in bases:
                if isinstance(base, ModelBase):
                    if base._meta.has_field('lockersmeta'):
                        raise TypeError, 'not implemented: user-defined lockersmeta'

        ncls = super(ModelBase, mcls).__new__(mcls, name, bases, attrs)

        ncls._meta.class_name = class_name

        return ncls

    def __repr__(cls):
        return '<model class %s version %s at %x>'%(cls._meta.class_name, cls._meta.class_version, id(cls))


class FactoryType(ModelBase):
    '''
    Factory metaclass.

    This metaclass can be instantiated directly to produce a new class,
    which can then be inherited from. (That'll cause magic to happen in
    LockersObject, but never mind now.)

    The factory is responsible for delegating to different versions of
    the target class jobs like loading and creating instances.

    You can also subclass Factory if you want to add more logic to loading
    or fetching. Look at User.py for an example.

    I use the variable "fac" here to refer to the Factory class, since "cls"
    refers to the class I'm instantiating/creating. FIXME?
    '''
    def __new__(mcls, *args, **kwargs):
        if len(args) == 3 and len(kwargs) == 0:
            # NOTE: this constructs a new ModelBase object, not a new
            # factory!  This is the "inheritance" case, someone is
            # trying to create a class that inherits from a Factory
            # *instance*.
            #
            # We don't want the new class to be an instance of the
            # Factory metaclass -- it's probably just a
            # ModelBase. (NOTE the use of Modelbase as the mcls
            # argument to the super call.)
            #
            # If they need some other metaclass, they need to specify
            # it themselves.
            #
            # kwargs are basically provided *just* to work around this
            # restriction
            return super(FactoryType, mcls).__new__(ModelBase, *args)
        if args:
            basename = args[0]
        else:
            basename = mcls.basename
        return super(FactoryType, mcls).__new__(mcls, basename + '__FAC', (object,),{})

    def __init__(fac, basename=None, *args):
        if basename:
            fac.basename = basename
        fac.classes = ModelBase.MODEL_CLASSES[fac.basename]
        fac._meta = fields.Options()
        fac._meta.multiclass = fac.classes.all_versions()
        fac._meta.class_name = fac.basename

    def isinstance(fac, inst):
        # TODO: change to __contains__?
        # FIXME: This is pretty horribly slow!
        if issubclass(inst.__class__, tuple(fac.classes.versions.values())):
            return True
        return False


class ModelFactory(FactoryType):
    def new(fac, t, node, *args, **kwargs):
        version = kwargs.pop('version', None)
        cls = fac.classes.get(version)
        o = cls.new(t, node, **kwargs)
        if version:
            o.version = version
        return o

    def existing(fac, t, node, *args, **kwargs):
        if not t.exists(node):
            raise DoesNotExistError, node

        want_version = kwargs.pop('as_version', None)
        migrate = kwargs.pop('migrate', True)

        verno = None
        disk_version = None
        if 'lockersmeta' in t.list_parts(node):
            meta = syck.load(t.get_part(node, 'lockersmeta'))
            disk_version = meta['version']
            verno = version_from_disk(disk_version)

        if not fac.classes.has_key(verno):
            raise TypeError, "couldn't find a class in %s to load %s %s"%(fac, verno, disk_version)
        cls = fac.classes.get(verno)
        o = cls.existing(t, node, *args, **kwargs)

        # migrate o
        if migrate:
            o = fac.classes.migrate(o, to=want_version)

        return o

    def __call__(fac, t, node, *args, **kwargs):
        return fac.existing(t, node, *args, **kwargs)

    def get(fac, t, node, *args, **kwargs):
        if t.exists(node):
            return fac.existing(t, node, *args, **kwargs)

        return fac.new(t, node, *args, **kwargs)

    def __repr__(fac):
        return '<factory for %s at %x>'%(fac.basename, id(fac))

class ChangesetFactory(FactoryType):
    def __call__(fac, *args, **kwargs):
        return fac.classes.get()(*args, **kwargs)

    def load(fac, t, rev):
        message = t.commit_message(rev)
        timestamp = t.commit_timestamp(rev)
        documents = message.split('\n---\n')  # FIXME: ughhhh
        message = documents[-1]

        serialized = syck.load(message)
        version = serialized['version']
        classname = version[0]['name']
        #classver = Migratable.version_from_disk(version[0]['version'])
        classholder = ModelBase.MODEL_CLASSES[classname]
        cls = classholder.get()  # FIXME: get the right version, and migrate

        changeset = cls(serialized)
        changeset.revision = rev
        changeset.timestamp = timestamp

        import lockers.user # FIXME?
        if hasattr(changeset, 'committer'):
            changeset.committer = lockers.user.UserObject.find_by_id(t, changeset.committer)
        else:
            changeset.committer = lockers.user.SystemUser

        return changeset

class CachedModelFactory(ModelFactory):
    def __init__(fac, basename, path=None):
        super(CachedModelFactory, fac).__init__(basename)
        fac._path = path
        fac._node = node.Node('cache', path, untracked=True)

    # FIXME: All these methods suck hardcore
    def __call__(fac, t, **kwargs):
        cache = super(CachedModelFactory, fac).__call__(t, fac._node, **kwargs)
        t.autosave(cache)
        return cache

    def get(fac, t, *args, **kwargs):
        obj = None
        if t.exists(fac._node):
            obj = super(CachedModelFactory, fac).existing(t, fac._node, *args, **kwargs)
        else:
            obj = super(CachedModelFactory, fac).new(t, fac._node, *args, **kwargs)

        t.autosave(obj)
        return obj

    def existing(fac, t, **kwargs):
        cache = super(CachedModelFactory, fac).existing(t, fac._node, **kwargs)
        t.autosave(cache)
        return cache

    def new(fac, t, **kwargs):
        cache = super(CachedModelFactory, fac).new(t, fac._node, **kwargs)
        t.autosave(cache)
        return cache

def Factory(name):
    print 'deprecated use of Factory for', name
    classes = ModelBase.MODEL_CLASSES.get(name)

    if issubclass(classes.get(), Changeset_1):
        return ChangesetFactory(name)
    return ModelFactory(name)

def version_from_disk(disk_version):
    return utils.list_to_tuple(version_from_disk_helper(disk_version))


def version_from_disk_helper(disk_version):
    version = []
    version.append(disk_version[0]['version'])
    for d in disk_version[1:]:
        version.append(version_from_disk_helper(d))

    return version

class Migratable(object):
    # FIXME: These methods need to be rewritten
    def version_to_disk(self):
        return self.version_to_disk_helper(self.version)

    @classmethod
    def version_to_disk_helper(cls, inst_version):
        version = []
        version.append({'name': cls._meta.class_name, 'version': inst_version[0]})
        for i, c in enumerate(cls._meta.parent_classes):
            version.append(c.version_to_disk_helper(inst_version[i+1]))

        return version

    @classmethod
    def compare_version_from_disk(cls, disk_version):
        if disk_version[0]['name'] != cls._meta.class_name:
            return False

        if disk_version[0]['version'] != cls._meta.class_version[0]:
            return False

        for i, d in enumerate(disk_version[1:]):
            if not cls._meta.parent_classes[i].compare_version_from_disk(d):
                return False

        return True


_fields = fields
class Changeset_1(Migratable):
    """Changeset class, corresponding to a commit message.

    If the Changeset class has a 'message' attribute, this will be shown
    specially in the git logs. If not, make_message() is called, and
    its value used in its stead.

    Changeset timestamps are recorded automatically, but user needs to
    be recorded by the developer."""
    __metaclass__ = ModelBase
    def __init__(self, fields={}, want_version=None, migrate=False,
                 new_version=None, **kwargs):
        self.__dict__.update(fields)  # FIXME: ?????
        super(Changeset_1, self).__init__(new_version=new_version, migrate=migrate,
                                      want_version=want_version)
        self.version = self._meta.class_version
        self.version = self.version_to_disk()
        # None catches any Parts -- shouldn't be any, for a Changeset.
        # There's no index part, really, either; but that's where all
        # the fields got collected to. FIXME?
        _fields.fill(None, fields, self,
                    from_=self._meta.get_field('index')._meta.fields)

    def make_message(self):
        return 'No user-provided message for %s'%(self,)

    def serialize(self):
        return yaml_dump(self.__dict__)

    def as_commit_message(self):
        if not hasattr(self, 'message') or not isinstance(self.message, (str, unicode)):
            message = self.make_message()
        else: message = self.message

        message = message + "\n\n" + self.serialize()

        return message

Changeset = ChangesetFactory('Changeset')

class LockersObject_1(Migratable):
    __metaclass__ = ModelBase

    # Streaming to disk is just a matter of saving a file-like object as part of a
    # LockersObject. The VCS will take care of it.

    class lockersmeta(fields.Part):
        version = fields.TupleField()
        enclosed_by = fields.NodeField()

    def __init__(self, t, node, create=False,
                 want_version=None, migrate=True, new_version=None,
                 want_fields=None,
                 **kwargs):
        '''

        '''
        self.__mine = {'node': node, 'transaction': t}
        if t.exists(node):
            if create == True:
                raise AlreadyExistsError, node

            self.__mine['is_new'] = False
            if kwargs:
                raise ValueError, 'cannot load node with keyword arguments %s'%kwargs

            parts = t.list_parts(node)
            if 'lockersmeta' in parts:
                meta_info = syck.load(t.get_part(node, 'lockersmeta'))
                disk_version = meta_info['version']

                self.version = version_from_disk(disk_version)

                if not self.compare_version_from_disk(disk_version):
                    raise TypeError, 'trying to load %s object with cls %s %s' % (
                        disk_version, self._meta.class_name, self._meta.class_version)

            RawPart = fields.RawPart

            if want_fields == None: want_fields = parts
            for partname in parts:
                # FIXME: partial loads including fields from index?
                if partname != 'index' and partname not in want_fields: continue

                cls = self._meta.get_field(partname, default=None)
                # FIXME: issubclass
                if isinstance(cls, RawPart) and cls.stream:
                    part = t.stream_part(node, partname)
                elif isinstance(cls, RawPart):
                    part = t.get_part(node, partname)
                else:
                    # Try to load YAML: fall back to text
                    try:
                        part = syck.load(t.get_part(node, partname))
                    except Exception, e:
                        print 'WARNING: exception when unserializing', node, partname, e
                        part = t.get_part(node, partname)

                    # Try to parse as Part: fall back to whatever we got
                    try:
                        cls = self._meta.get_field(partname) # FIXME: again?
                        part = cls.convert(part, self, partname)
                    except fields.FieldDoesNotExist, e:
                        pass

                setattr(self, partname, part)

        else:
            if create == False:
                raise DoesNotExistError, node

            self.__mine['is_new'] = True
            # fields should always show up wrapped in an index Part class
            # -- done automatically by fields.FieldCollector, so pass a
            # malicious None here to catch unwrapped fields in the class.
            fields.fill(self, kwargs, None)

            # Passing None here too: this might cause problems if we ever
            # allow sub-parts.
            fields.fill(None, kwargs, self.index,
                        from_=self._meta.get_field('index')._meta.fields)

        if hasattr(self._meta, 'class_version'):
            self.version = self._meta.class_version

        super(LockersObject_1, self).__init__(want_version=want_version, new_version=new_version, migrate=migrate)

    def set_version(self, newv):
        self.lockersmeta.version = newv

    def get_version(self):
        return self.lockersmeta.version
    version = property(get_version, set_version)

    def compare_version(self, disk_version):
        return disk_version[-1]['version'] == self._meta.class_version

    def is_new(self):
        return self.__mine['is_new']

    def is_dummy(self):
        return isinstance(self.__mine['transaction'], DummyTransaction)

    def save(self, changeset, **kwargs):
        if self.__mine['node'].revision:
            raise AlterPastError, 'cannot save an object based on revision %s'%\
                self.__mine['node'].revision

        # Is this actually useful? Doubtful.
        if self.version == False:
            raise BackendException, 'cannot save unversioned object'

        t = self.__mine['transaction']
        node = self.__mine['node']

        # FIXME: wipe old parts in the node

        for partname in self.__dict__:
            if partname.startswith('_'): continue
            # FIXME: this sucks, but we have weird race conditions in the
            # Session objects where the "raw" version attribute gets saved
            # and then read before the "cooked" version_to_disk comes out.
            # This will most likely cover those problems, but Session doesn't
            # seem to be sufficiently reentrant..
            if partname == 'lockersmeta':
                d = dict(self.lockersmeta.__dict__)
                d['version'] = self.version_to_disk()
                t.save_part(node, partname, yaml_dump(d))
                continue

            if self.__dict__[partname] == PRESERVE: continue
            cls = self._meta.get_field(partname, default=None)
            if isinstance(cls, fields.RawPart):
                t.save_part(node, partname, self.__dict__[partname])
            else:
                t.save_part(node, partname, yaml_dump(self.__dict__[partname]))

        for partname in kwargs:
            # FIXME, FIXME, hack hack hack
            # I have a thing that saves files, can't stream from it, so need to
            # call a lambda that will create the file, then add it
            cls = self._meta.get_field(partname, default=None)
            if isinstance(cls, fields.RawPart):
                t.save_part(node, partname, kwargs[partname])


        #print "Saving an object " + self.__class__.__name__, yaml_dump(self.__class__.__name__)


        if not isinstance(changeset, Changeset_1):
            # FIXME: This should really not be here.
            # We should really enforce that this is a Changeset object
            if hasattr(self, 'COMMENT'): changeset = self.COMMENT(fields=changeset)
            else:
                print "WARNING: constructing default Changeset for", changeset
                changeset = Changeset(fields=changeset)


        if not node.untracked:
            # Only act on changed dependencies if the target is
            # tracked.  Otherwise we get into loops with saving parts
            # of the CacheDependency needing to save the CacheDependency.
            lockers.indexes.Indexes(t).was_changed(self.get_node())

            t.commit(changeset.as_commit_message())

    def __setattr__(self, prop, val):
        '''A complicated check to see whether we should allow this set.

        We allow sets to replace existing parts of ours, and we allow sets
        that get passed to our index part.

        We also allow any sets starting with _, since we don't save them
        (I think).

        We also allow sets on undefined parts if we have an
        add_extra_fields attribute. I think this might be a mistake.'''
        if prop.startswith('_') or self._meta.has_field(prop) or \
                getattr(self, 'add_extra_fields', False) or \
                self._meta.get_field('index')._meta.has_field(prop) or \
                isinstance(getattr(self.__class__, prop, None), property):
            return super(LockersObject_1, self).__setattr__(prop, val)
        raise TypeError, "'%s' object does not have a part '%s'"%(self.__class__.__name__, prop)

    def get_node(self):
        return self.__mine['node']

    def get_transaction(self):
        return self.__mine['transaction']

    @classmethod
    def new(cls, *args, **kwargs):
        '''Create a new LockersObject for this node.

        Raises AlreadyExistsError if this node already exists.

        Equivalent to calling this class's constructor with create=True.'''
        kwargs.setdefault('create', True)
        return cls(*args, **kwargs)

    @classmethod
    def existing(cls, *args, **kwargs):
        '''Get an existing LockersObject for this node.

        Raises DoesNotExistError if this node doesn't exist.

        Since the LockersObject constructor defaults to this mode, you
        don't have to call this classmethod explicitly.'''
        kwargs.setdefault('create', False)
        return cls(*args, **kwargs)

    @classmethod
    def get(cls, *args, **kwargs):
        '''Get a LockersObject for this node, regardless of existence.

        Load an existing one if any, otherwise, create an empty one.

        Equivalent to calling this class's constructor with create=None.'''
        kwargs.setdefault('create', None)
        return cls(*args, **kwargs)

    @classmethod
    def copy(cls, o):
        # Used by ModelHolder.migrate to update the class of an existing object
        t = o.__mine['transaction']
        node = o.__mine['node']

        dct = dict(o.__dict__)
        for key in dct.keys():
            if key.startswith('_'): del dct[key]

        new_o = cls.get(t, node, **dct)
        return new_o

LockersObject = ModelFactory('LockersObject')
class UntrackedObject_1(LockersObject):
    def __init__(self, t, node, *args, **kwargs):
        if not node.untracked:
            print "WARNING: Creating untracked object from node without untracked attribute (%s)" % node
            node = node.replace(untracked=True)

        super(UntrackedObject_1, self).__init__(t, node, *args, **kwargs)

    def save(self, **kwargs):
        super(UntrackedObject_1, self).save(Changeset({'message': 'I don\'t exist'}), **kwargs)

    def register_dependencies(self, targets):
        lockers.indexes.Indexes(self.get_transaction()).register_dependencies(self.get_node(), targets)

UntrackedObject = ModelFactory('UntrackedObject')

# FIXME: importing madness
import lockers.indexes
