# Node.py -- a representation of a user request
import os.path
import copy
import cgi
from paste.util.multidict import MultiDict
from request import LockersRequest
from url import URL

class LockersNode(object):
    parts = ['module', 'path', 'revision', 'untracked']
    def __init__(self, *args, **kwargs):
        # forms:
        # - LockersRequest, in which case "figure it out"
        # - Node, in which case copy
        # - args are a prefix of parts: Node('wiki')
        # - kwargs specifies different parts to replace with
        def _fail():
            raise TypeError, "unknown location format"

        self.module = ''
        self.path = ''
        self.revision = None
        self.untracked = False

        if len(args) == 1 and isinstance(args[0], LockersRequest):
            self._construct_from_request(args[0])
        elif len(args) == 1 and isinstance(args[0], self.__class__):
            self._construct_from_node(args[0])
        elif len(args) > 0:
            if len(args) >= 1:
                self.module = args[0]
            if len(args) >= 2:
                self.path = args[1]
            # other args are weird
            if len(args) >= 3:
                raise TypeError, "unused arg in Node %s"%(tuple(args),)

        self._update(kwargs)

        if self.path and self.path.startswith('/'):
            self.path = self.path[1:]

    def __reduce_ex__(self, protocol):
        return (Node, (), {'module': self.module, 'path': self.path, 'revision': self.revision, 'untracked': self.untracked})

    def __hash__(self):
        return hash(''.join([self.module, '/', self.path,
                             '?', str(self.revision), ':', str(self.untracked)]))

    def _construct_from_request(self, request):
        self.module = request.script_name.strip('/')
        self.path = request.path_info
        if request.GET.has_key('revision'):
            self.revision = request.GET['revision']

        # FIXME: this sucks; we only need it for base_url
        self.request = request

    def _construct_from_node(self, node):
        self.module = node.module
        self.path = node.path
        self.revision = node.revision
        self.untracked = node.untracked

    def _update(self, dct):
        if dct.has_key('module'):
            self.path = ''
            self.module = dct.pop('module')
            self.untracked = False
            # self.revision stays

        if dct.has_key('path'):
            self.untracked = False
            self.path = dct.pop('path')
            # self.revision stays

        if dct.has_key('revision'):
            self.revision = dct.pop('revision')
            self.untracked = False

        if dct.has_key('untracked'):
            self.untracked = dct.pop('untracked')

        if dct: raise ValueError, "unused keyword arguments: %s" %dct

    def __eq__(self, rhs):
        return type(self) == type(rhs) and self.module == rhs.module and self.path.strip('/') == rhs.path.strip('/') and \
            self.revision == rhs.revision

    def __cmp__(self, rhs):
        # FIXME: completely wrong comparison for revision
        # FIXME: test this??
        if self.revision < rhs.revision: return -1
        if self.revision > rhs.revision: return 1
        if self.module < rhs.module: return -1
        if self.module > rhs.module: return 1
        if self.path < rhs.path: return -1
        if self.path > rhs.path: return 1
        return 0

    def __repr__(self):
        return 'Node(%s, %s, **%s)'%(repr(self.module), repr(self.path),
                                     repr({'revision': self.revision, 'untracked': self.untracked}))

    def url(self):
        return self.extend_url(URL(''))

    def extend_url(self, url):
        u = url
        for p in self.url_components():
            u = u[str(p)]
        return u

    def base_url(self):
        assert hasattr(self, 'request')
        if self.request.environ.has_key('ORIG_SCRIPT'):
            u = URL('http://%s/'%self.request.environ['HTTP_HOST'])[self.request.environ['ORIG_SCRIPT']]
        else: u = URL('')
        return u

    def push(self, module):
        # FIXME: either rename or make it a mutator
        new = Node(self)
        new.path = os.path.join(self.module, self.path)
        new.module = module
        return new

    def url_to(self, *args, **kwargs):
        if len(args) == 0:
            target = None
        elif len(args) == 1:
            target = args[0]
        else:
            raise TypeError, 'too many arguments to Node.url_to'

        newdct = {}
        if target:
            for key in target.parts:
                val = getattr(target, key)
                # FIXME? What if someone wants to explicitly remove the module?
                # Does that even make sense? Probably not.
                if val:
                    newdct[key] = val

        for key in self.parts:
            if kwargs.has_key(key):
                newdct[key] = kwargs.pop(key)
        n = self.replace(**newdct)
        u = n.extend_url(self.base_url())
        for key in kwargs:
            u = u["%s=%s"%(key, kwargs[key])]
        return u.href

    def url_components(self):
        ret = [self.module, self.path]
        if self.revision:
            ret.append('revision=%s'%self.revision)
        # untracked has no place in URLs
        if self.untracked:
            print "Warning: url_components taken of %s"%self
        return ret

    def filename(self, part=None):
        """The corresponding location in the filesystem for this page.

        Typically we want this to be a directory, in case we decide to store
        the components of this object in more than one file. We want this 
        directory to contain nothing *but* components of this object, so we can
        do a 'git log' to see what changed in this object (and not its 
        children).

        The filename of one component of the object is available using the 
        'part' parameter. The default component right now is called 'index',
        so any object is stored at path/index/index. I know, this sucks."""

        path = os.path.join(self.module, self.path, 'index')
        if part:
            return os.path.join(path, part)
        return path

    def part(self, part='index'):
        '''Convenience method for self.filename(part), with a mandatory part
        (defaults to index).'''
        return self.filename(part)

    def location(self, part=''):
        '''Return the parent directory of the node, i.e. where the
        node's children would be.'''
        return os.path.join(self.module, self.path)

    def clone(self, *args, **kwargs):
        # create a new node like this one, except with given args/kwargs
        # and "smart replacing"
        return self.__class__(self, *args, **kwargs)

    def replace(self, **dct):
        return self.clone(**dct)

    def child(self, *paths):
        oldpath = self.path.strip('/')
        newpath = '/'.join([oldpath]+[path.strip('/') for path in paths])
        return self.clone(path=newpath)

    def parent(self):
        """Return a new node which represents the file's parent file."""
        if self.path == '/':
            print 'Warning: computing parent of module root node %s'%self
            return self.clone()
        names = self.path.split('/')[:-1]
        return self.clone(path='/'.join(names))

    def last_path_part(self):
        return self.path.split('/')[-1]

    def append(self, path):
        self.path = os.path.join(self.path, path)

Node = LockersNode
