import math

class Paginator(object):
    '''
    Paginator.

    FIXME: not integrated at all with querying. Since most queries are
    cheap, I don't care too much.

    Paginates based on "elements skipped", not "page the user is
    on". "Page" seems more aesthetic than "skip" until you start to
    consider things like: what if user has already seen pages 1-4 and
    now wants to change the number-per-page to 13? What if one user
    sends a link to another one with a bigger number-per-page?

    We reimplement how Google does their pagination. They use
    skip-based pagination, and the next and previous buttons always go
    to the next result over, but their direct-page-links ignore your
    current skip. This leads to some weird results like "I'm on page 3
    of 3 but there's a Next button".

    After some thought, I've decided this is probably the best course
    of action (when supported by BIG FRIENDLY previous/next buttons in
    the web design) and besides, Google's probably done research into
    this and I don't feel like reinventing the wheel.
    '''
    def __init__(self, node, lst, num_per_page, skip=0, window_size=5, user_request=None):
        '''Computes a bunch of useful pagination things.

        Designed to be fed to a Django template. Calculates the page
        number based on current skip count.

        @param num_per_page: default number of elements to include per
            page. This can be overridden by user_request.

        @param window_size: number of pages to have in a window --
            centered around the page we're currently on.
            If window_size is even, we'll freak out.

        @param user_request: a dictionary of parameters from the
            user. We read certain specific overrides (num, skip) from
            this dictionary and convert them to int. The expected
            input is request.GET, but any dict will work.
        '''
        self.node = node
        self.objects = lst
        self.num_per_page = num_per_page
        # 0 <= skip < len(lst)
        self.skip = skip

        if user_request:
            if 'num' in user_request:
                self.num_per_page = int(user_request['num'])
            if 'skip' in user_request:
                self.skip = int(user_request['skip'])

        # page is CALCULATED, not controlled by developer!
        self.page = int(math.ceil(float(self.skip) / self.num_per_page)+1)

        self.window_size = window_size
        if self.window_size % 2 == 0:
            raise TypeError, 'window_size cannot be even'

        # Window computation done here because it all kinda has to be done
        # at once
        self.window_half = int(math.floor(self.window_size-1)/2)
        start = self.page - self.window_half
        end = self.page + self.window_half  # inclusive!
        if start < 1:
            end += 1 - start
            start = 1
        if end > self.number_of_pages():
            end = self.number_of_pages()

        self.window_inclusive = [start, end]

    def number_of_pages(self):
        n = len(self.objects)
        pfrac = float(n)/self.num_per_page
        return int(math.ceil(pfrac))

    def at_beginning(self):
        return self.skip == 0

    def at_end(self):
        return self.skip + self.num_per_page >= len(self.objects)

    def is_viewing_all(self):
        return self.at_beginning() and self.at_end()

    def has_previous_page(self):
        return not self.first_page()

    def has_next_page(self):
        return not self.last_page()

    def _make_page(self, page, newskip):
        '''Factored out, in case Page turns out to need the number of objects
        per page or other members of the Paginator class'''
        return Page(self.node, page, newskip, num=self.num_per_page)

    def previous_page(self):
        # We don't do bounds checking using is_first_page or is_last page
        start = self.skip-self.num_per_page
        if start < 0:
            start = 0
        return self._make_page(self.page-1, start)

    def next_page(self):
        return self._make_page(self.page+1, self.skip+self.num_per_page)

    def first_page(self):
        return self._make_page(1, 0)

    def last_page(self):
        return self._make_page(self.number_of_pages(), (self.number_of_pages()-1)*self.num_per_page)

    def pages_before(self):
        start, end = self.window_inclusive
        #skipstart = self.skip - (self.page-start)*self.num_per_page
        skipstart = (self.page-1) * self.num_per_page
        lst = []
        n = skipstart
        for page in xrange(self.page-1, start-1, -1):
            n -= self.num_per_page
            lst.append(self._make_page(page, n))
        lst.reverse()
        return lst

    def pages_after(self):
        start, end = self.window_inclusive
        lst = []
        n = (self.page-1) * self.num_per_page
        for page in xrange(self.page+1, end+1):
            n += self.num_per_page
            lst.append(self._make_page(page, n))
        return lst

    def window_includes_first_page(self):
        start, end = self.window_inclusive
        return start == 1

    def window_includes_last_page(self):
        start, end = self.window_inclusive
        return end == self.number_of_pages()

    def slice(self):
        start = self.skip
        return self.objects[start:(start+self.num_per_page)]

class Page(object):
    '''Helper object to hold both number and link for Paginator'''
    def __init__(self, node, pagenum, skip, **kwargs):
        self.node = node
        self.skip = skip
        self.pagenum = pagenum
        self.kwargs = kwargs

    def url(self):
        return self.node.url_to(skip=self.skip, **self.kwargs)

    def number(self):
        return self.pagenum

    def __eq__(self, rhs): # mostly for debugging
        return self.node == rhs.node and self.skip == rhs.skip and \
            self.pagenum == rhs.pagenum

    def __repr__(self):
        return "Page(%r, %r, %r)" % (self.node, self.pagenum, self.skip)
