from __future__ import with_statement
import os
import sys
import time
import md5
import random
import copy
from webob import Response
from django.utils.http import cookie_date
from models.models import UntrackedObject
from models.fields import DictPart, StrPart
from node import Node
from application import LockersApplication, CONFIG

SESSION_COOKIE_NAME = 'sessionid'
SESSION_COOKIE_AGE = 60 * 60 * 24 * 7 * 2
SESSION_COOKIE_PATH = '/'
SESSION_COOKIE_SECURE = False


def _new_session_key():
    try:
        pid = os.getpid()
    except AttributeError:
        # No getpid() in Jython, for example
        pid = 1
    return md5.new("%s%s%s%s" % (random.randint(0, sys.maxint - 1),
                                 pid, time.time(), CONFIG['secret_key'])).hexdigest()

class Session_1(UntrackedObject):
    index = DictPart()
    session_key = StrPart()

    def __init__(self, *args, **kwargs):
        session_key = kwargs.pop('session_key')
        super(Session, self).__init__(*args, **kwargs)

        # on init, session_key is the empty string
        if not getattr(self, 'session_key', False):
            self.session_key = session_key

    def get_expire_at_browser_close(self):
        return False

    def get_expiry_age(self):
        return SESSION_COOKIE_AGE

Session = Session_1

class SessionMiddleware(LockersApplication):
    def __init__(self, app):
        super(SessionMiddleware, self).__init__()
        self.app = app

    def __call__(self, request):
        # Note that we do two transactions here -- this might open us
        # up to a weird race where two requests come in from the
        # browser at the same time and both modify the session and
        # only the last one gets written.  I really doubt there's much
        # I can do about this at this level of abstraction. The only
        # really correct way to deal with this is to prevent serving
        # more than one request from the same browser at once. We could also
        # try to do some kind of merge, but without Git smarts, that's hard,
        # and to use Git smarts we'd have to commit on every session change..
        # bad mojo.
        session_key = request.COOKIES.get(SESSION_COOKIE_NAME, None)
        if not session_key: session_key = _new_session_key()
        n = Node(module='sessions', path='/'+session_key, untracked=True)

        with self.backend.new_transaction() as t:
            session = Session.get(t, n, session_key=session_key)
            request.session = session.index

        session_old = copy.deepcopy(request.session)

        value = self.app(request)
        if not value:
            print 'WARNING: SessionMiddleware application did not return anything', self.app
            value = Response('')

        with t:
            # The commented-out if statement will force a save every
            # request, whether dirty or not. The problem is that
            # accessing Git is still really not re-entrant, so if a
            # browser requests twenty images at once, you get a flurry
            # of saving/reloading sessions, including some reloading
            # of partially saved sessions (with 'version' or 'index'
            # parts not completely serialized or otherwise badly
            # screwed up).

            # Since a session is almost always only going to get
            # changed once in each user action, it probably makes more
            # sense to do a recursive check. The only danger is that
            # some user type will not get copied correctly, or not
            # compare equally somehow, and we'll end up re-saving the
            # session even though nothing's changed, or not saving it
            # when something has changed.

            #if request.session or session_old:

            if request.session != session_old:
                if session.get_expire_at_browser_close():
                    max_age = None
                    expires = None
                else:
                    max_age = session.get_expiry_age()
                    expires_time = time.time() + max_age
                    expires = cookie_date(expires_time)
                session.save()

                value.set_cookie(SESSION_COOKIE_NAME, session.session_key,
                                 max_age=max_age, expires=expires,
                                 path=SESSION_COOKIE_PATH,
                                 secure=SESSION_COOKIE_SECURE or None)
        return value
