import os
import sys
import imp

class FileApp(object):
    def __init__(self, filename):
        self.filename = filename

    def __call__(self, request):
        return Response(file(self.filename))

class URLParser(object):
    '''A simplistic URLParser which is an extremely stripped down version of 
    the paste URLParser. We did this to use WebOb's request in a more native
    way.
    '''

    parsers_by_directory = {}

    global_constructors = {}

    def __init__(self, global_conf, directory, base_python_name,
                 hide_extensions=None,
                 ignore_extensions=None,
                 constructors=None,
                 index_names=None):
        '''hide_extensions: these cannot be looked up even with an exact name
        ignore_extensions: these can be looked up with an exact name, but
            partial matches will be ignored'''
        # Not implemented: index_files, constructor_conf
        self.directory = directory
        self.base_python_name = base_python_name

        if hide_extensions == None:
            hide_extensions = global_conf.get(
                'hide_extensions', ('.pyc', '.bak', '.py~', '.pyo'))
        self.hide_extensions = hide_extensions

        if ignore_extensions == None:
            ignore_extensions = global_conf.get(
                'ignore_extensions', ())
        self.ignore_extensions = ignore_extensions

        if index_names == None:
            index_names = global_conf.get(
                'index_names', ('index', 'Index'))
        self.index_names = index_names

        self.constructors = self.global_constructors.copy()
        if constructors:
            self.constructors.update(constructors)

    def __call__(self, request):
        application, filename = self.find_application(request)
        return application(request)

    def find_application(self, request):
        if request.path_info == '/' or request.path_info == '':
            # try index files
            for index_name in self.index_names:
                filename = self.find_file(request, index_name)
                if filename: break
            else:
                filename = None

        else:
            name = request.path_info_pop()
            filename = self.find_file(request, name)

        if filename is None:
            raise NotImplemented, "couldn't find file"

        else:
            return self.get_application(request, filename), filename

    def find_file(self, request, base_filename):
        possible = []
        directory = self.directory
        for filename in os.listdir(directory):
            base, ext = os.path.splitext(filename)
            full_filename = os.path.join(directory, filename)
            if not base: continue
            if ext in self.hide_extensions: continue

            if filename == base_filename:
                possible.append(full_filename)
                continue

            if ext in self.ignore_extensions: continue
            if base == base_filename:
                possible.append(full_filename)
                continue

        if len(possible) > 1:
            # There might be an exact match and an inexact match
            if os.path.join(directory, base_filename) in possible:
                return os.path.join(directory, base_filename)

            else:
                raise RuntimeError, \
                    "Ambiguous URL: %s; too many matches %s"%(request.script_name, ', '.join(possible))
        if len(possible) == 0:
            raise RuntimeError, \
                "Couldn't find anything for %s in %s" % (base_filename, directory)
        return possible[0]

    def get_application(self, request, filename):
        if os.path.isdir(filename):
            t = 'dir'
        else:
            t = os.path.splitext(filename)[1]
        constructor = self.constructors.get(t, self.constructors.get('*'))
        if constructor is None:
            raise RuntimeException, 'No constructor found for %s\n'%t
        app = constructor(self, request, filename)
        if app is None:
            raise RuntimeError, "Constructor didn't return app for %s" % filename
        return app

    def register_constructor(cls, extension, constructor):
        """
        Register a function as a constructor.  Registered constructors
        apply to all instances of `URLParser`.

        The extension should have a leading ``.``, or the special
        extensions ``dir`` (for directories) and ``*`` (a catch-all).

        `constructor` must be a callable that takes two arguments:
        ``request`` and ``filename``, and returns a WSGI application.
        """
        d = cls.global_constructors
        assert not d.has_key(extension), (
            "A constructor already exists for the extension %r (%r) "
            "when attemption to register constructor %r"
            % (extension, d[extension], constructor))
        d[extension] = constructor
    register_constructor = classmethod(register_constructor)

    def get_parser(self, filename, base_python_name):
        try:
            return self.parsers_by_directory[(filename, base_python_name)]
        except KeyError:
            parser = self.__class__(
                {},
                filename, base_python_name,
                #index_names=self.index_names,
                hide_extensions=self.hide_extensions,
                ignore_extensions=self.ignore_extensions,
                constructors=self.constructors)
            self.parsers_by_directory[(filename, base_python_name)] = parser
            return parser


def make_directory(parser, request, filename):
    base_python_name = parser.base_python_name
    if base_python_name:
        base_python_name += "." + os.path.basename(filename)
    else:
        base_python_name = os.path.basename(filename)
    return parser.get_parser(filename, base_python_name)

URLParser.register_constructor('dir', make_directory)

def make_unknown(parser, request, filename):
    return FileApp(filename)

URLParser.register_constructor('*', make_unknown)


def load_module(request, filename, base_module_name):
    module_name = os.path.splitext(os.path.basename(filename))[0]
    if base_module_name:
        module_name = base_module_name + '.' + module_name
    return load_module_from_name(filename, module_name,
                                 request.environ['wsgi.errors'])

def load_module_from_name(filename, module_name, errors):
    if sys.modules.has_key(module_name):
        return sys.modules[module_name]
    fp = None
    if '.' in module_name:
        parent_name = '.'.join(module_name.split('.')[:-1])
        base_name = module_name.split('.')[-1]
        parent = load_module_from_name(os.path.dirname(filename),
                                       parent_name, errors)
    else:
        base_name = module_name
    fp = None
    try:
        fp, pathname, stuff = imp.find_module(
            base_name, [os.path.dirname(filename)])
        module = imp.load_module(module_name, fp, pathname, stuff)
    finally:
        if fp is not None:
            fp.close()
    return module

def make_py(parser, request, filename):
    module = load_module(request, filename, parser.base_python_name)
    if not module:
        return None
    # FIXME: weird webob thing; I never used this
    #if hasattr(module, 'application') and module.application:
    #    return getattr(module.application, 'wsgi_application', module.application)
    base_name = module.__name__.split('.')[-1]
    if hasattr(module, base_name):
        obj = getattr(module, base_name)
        if hasattr(obj, 'wsgi_application'):
            return obj.wsgi_application
        else:
            # @@: Old behavior; should probably be deprecated eventually:
            return getattr(module, base_name)()
    request.environ['wsgi.errors'].write(
        "Cound not find application or %s in %s\n"
        % (base_name, module))
    return None

URLParser.register_constructor('.py', make_py)
