import hashlib
import random
import time
from models import fields, models
import node
from request import LockersRequest

SALT = 'abcde'              # FIXME: generate randomly once per Lockers install

def get_hexdigest(algo, *args):
    h = hashlib.new(algo)
    for a in args: h.update(a)
    return h.hexdigest()

def round_time(time, steps=0):
    '''time is in seconds-since-the-epoch'''
    return time-(time%(7*60)) # last seven-minutes-ago

def challenge():
    '''
    It's kinda stupid to worry about replay attacks, since anyone can just
    copy a session cookie and pretend to be someone else. But I'd rather this
    not be the weakest link.
    '''
    return get_hexdigest('sha1', str(round_time(time.time())))

def is_challenge_recent(c):
    # compute a few challenges, and see if c matches any of them
    t = round_time(time.time())
    for i in range(2):
        if c == get_hexdigest('sha1', str(t)): return True
        t -= 7*60 # from the previous seven-minutes-ago
    return False


class UserObject_1(models.LockersObject):
    username = fields.StrField()
    realname = fields.StrField()
    email = fields.EmailField()
    # Note: not actually strings! ModelBase could get unhappy with us here
    secret = [fields.StrField()]  # could be OpenID, pw hash, public key id..
    groups = [fields.StrField()]
    permissions = [fields.StrField()]

    def parse(self, pw):
        return pw

    def unparse(self, secret):
        return secret

    def set_password(self, pw):
        # We force algo = 'sha1' here because I don't want to be smart enough
        # to make the login script hash things differently on the client side
        algo = 'sha1'
        # overwrites all other secrets: bad
        salt = SALT
        hash = get_hexdigest(algo, salt, pw)
        unparsed = self.unparse({'type': 'hash', 'hashtype': algo, 
                                 'hash': hash})
        
        self.index.secret = [unparsed]

    def check_password(self, pw):
        for s in self.index.secret:
            secret = self.parse(s)
            h = get_hexdigest(secret['hashtype'], SALT, pw)
            if secret['hash'] == h:
                return True

        return False

    def check_hashword(self, pw, challenge):
        # pw = hash(hash(salt + u_pw) + challenge)
        if not is_challenge_recent(challenge): return False
        for s in self.index.secret:
            secret = self.parse(s)
            h = get_hexdigest('sha1', secret['hash'], challenge)
            if pw == h:
                return True

        return False

    def is_authenticated(self):
        return True

    def userid(self):
        return self.get_node().last_path_part()

class UserFactory(models.ModelFactory):
    basename = 'UserObject'
    def get(self, *args, **kwargs):
        if len(args) >= 2:
            if isinstance(args[1], LockersRequest):
                return self.get_request(*args, **kwargs)
        return super(UserFactory, self).get(*args, **kwargs)

    def get_request(self, t, request):
        if request.session.has_key('userid'):
            n = node.Node('users', request.session['userid'])
            return super(UserFactory, self).get(t, n)
        return self.login(request, t)

    def login(self, request, t):
        username = request.POST.get('username', None)
        password = request.POST.get('password', None)
        hashword = request.POST.get('hashword', None)
        if username and (password or hashword):
            user = self.find_user(t, {'username': username})
            if len(user) > 1:
                print 'many users with the same username:', username, user
            user = user[0]

            if password:
                # Unhashed version; JS off in the browser?
                if user.check_password(password):
                    return self._success(request, user)

            if hashword:
                challenge = request.session.pop('challenge')
                if user.check_hashword(hashword, challenge):
                    return self._success(request, user)



        # FIXME: new anonymous user every time?
        return AnonymousUser.new(t, node.Node('users', '/anonymous/%09d'%1))

    # FIXME: real query methods on Backend
    def find_user(self, t, q):
        results = []
        for userid in t.listdir('users'):
            n = node.Node('users', userid)
            user = self.get(t, n)
            # FIXME: use properties here
            for key in q:
                if getattr(user.index, key) != q.get(key):
                    break
            else:
                results.append(user)

        return results

    def find_by_id(self, t, id):
        n = node.Node('users', id)
        return self.get(t, n)

    def _success(self, request, user):
        request.session['userid'] = user.get_node().path
        return user

def logout(request):
    del request.session['userid']

UserObject = UserFactory()


class AnonymousUser_1(models.LockersObject):
    ip = fields.StrField()
    def is_authenticated(self):
        return False

AnonymousUser = models.ModelFactory('AnonymousUser')

# FIXME: not a real object; can't be saved, doesn't exist?
class SystemUser_1(object):
    username = 'system'
    def userid(self):
        return '%09d'%(0,)

SystemUser = SystemUser_1()
