# VCS.py: the interface for a version control system for the Lockers
#   framework
import datetime
import types
import os
import lock
import utils
import errno
from utils import pexecl, pexecl_stream
from lockers.node import Node

class VCSException(Exception):
    pass

def revision_name(revid):
    if revid == None:
        return 'current revision'
    return 'revision %s'%revid

class FileNotFoundError(VCSException):
    def __init__(self, filename, revid):
        msg = "File %s does not exist in %s"%(filename, revision_name(revid))
        super(FileNotFoundError, self).__init__(msg)

# Are these objects needed? Right now we just use strings.
class RevisionId(object):
    def __init__(self, revid):
        self.revid = revid

    def next(self):
        pass

    def previous(self):
        pass

class Revision(object):
    def __init__(self, revid):
        self.revid = revid

    def commit_message(self):
        pass

    def changes(self):
        pass

class VCS(object):
    '''Controller for the version control system.

    The architecture of Lockers is such that the current directory is always
    the root of the version control system.

    Operations always take filenames.'''
    def __init__(self):
        '''Create a VCS interface. '''
        pass

    @classmethod
    def create(cls):
        '''Initialize the repository.

        This assumes that we're in the toplevel of the yet-to-be-created
        repo.'''
        cls._create()
        return cls()

    def is_new(self):
        '''Override if you need something fancier'''
        return self._new

    # Helper functions to assert preconditions
    def _no_node(self, thing):
        if isinstance(thing, Node):
            raise TypeError, 'VCS takes only filenames'

    def _exists_in_rev(self, filename, rev):
        self._no_node(filename)
        if not self.exists(filename, rev):
            raise FileNotFoundError, (filename, rev)

    def coerce_filename(self, filenames):
        '''Utility function to coerce a bunch of nodes using node.filename().'''
        filenames = list(filenames)
        for i in range(len(filenames)):
            if isinstance(filenames[i], Node):
                # filenames[i] = filenames[i].filename()
                raise TypeError, 'VCS only takes filenames'
        return filenames


    # QUERY FUNCTIONS START HERE.

    def current_revision(self, file=None):
        return self._current_revision(file)

    def patch(self, id):
        # returns a Changeset object? Takes a Changeset object?
        return self._patch(id)

    def get(self, filename, revid=None):
        '''Get the contents of a file as of some version.'''
        # returns the contents of the file as of revid
        self._exists_in_rev(filename, revid)
        return self._get(filename, revid)

    def stream(self, filename, revid=None):
        '''Get the contents of a file as of some version as a stream.

        A stream is a lazily-loaded iterator of file contents. You might do:

        f = vcs.stream('some_file')
        for chunk in f:
            print chunk

        The stream is loaded lazily so that when a RawPart(stream=True) is loaded,
        it doesn't necessarily create a zombie process.
        Reading fully from the stream is necessary to avoid zombie processes,
        so the intent is that processes are only created when the file is
        actually about to be read.'''
        self._exists_in_rev(filename, revid)
        return self._stream(filename, revid)

    def exists(self, filename, revid=None):
        '''Check whether a file existed in a given version.'''
        self._no_node(filename)
        try:
            self._get(filename, revid)
            return True
        except VCSException, e:
            return False

    def commit_message(self, revid):
        return self._commit_message(revid)

    def commit_timestamp(self, revid):
        timestring = self._commit_timestamp(revid)
        ts = self._unparse_timestamp(timestring)
        return ts

    def history(self, filenames, rev_from=None, rev_to=None, max=None):
        '''Returns all the revids which have new versions of files in filenames.

        Filenames is a sequence of files. Revisions returned contain changes
        to any of those files.

        @param rev_from: return only revisions further forward than this
        @param rev_to: return no revisions further forward than this
        @param max: return no more revisions than this
        @returns results in reverse-chronological order'''
        filenames = self.coerce_filename(filenames)
        return self._history(filenames, rev_from, rev_to, max)

    def listdir(self, directory='', revid=None, sort=True):
        # returns the contents of the directory as of revid
        # return value: list of filenames?
        lst = self._listdir(directory, revid)
        if sort: lst.sort()
        return lst

    def isdir(self, filename, revid=None):
        return self._isdir(filename, revid)

    # UPDATE FUNCTIONS START HERE.

    def dump(self, filename, contents):
        '''Dumps content into a file.'''
        ### FIXME -- this really needs to be thought out
        ### Should this really blindly create directories?
        ### It's still a bit cavalier, but I think it's like an ordinary "save"
        ### in some other DB.
        try:
            os.makedirs(os.path.dirname(filename))
        except:
            pass # don't care
        outf = file(filename, 'w')
        if hasattr(contents, 'read'):
            outf.write(contents.read())
            # NamedTemporaryFile deletes on close
            contents.close()
        elif isinstance(contents, types.FunctionType):
            # FIXME: Hack for PIL
            contents(filename)
        else:
            outf.write(contents)


    def add(self, files):
        '''Mark a version of a file as part of the next commit.'''
        if isinstance(files, (str, unicode)): files = [files]
        files = self.coerce_filename(files)
        return self._add(files)

    def commit(self, files, message):
        '''Create a new commit.

        files is a list of files or one file, or an empty list or None if
        all added files should be committed.'''
        if isinstance(files, str): files = [files]
        files = self.coerce_filename(files)
        return self._commit(files, message)

    def move(self, src, dest):
        if isinstance(src, str): src = [src]
        src = self.coerce_filename(src)
        return self._move(src, dest)

    def remove(self, filename):
        return self._remove(filename)

    def diff(self, src, dest, fname=None):
        return self._diff(src, dest, fname)

GITPATH = '/usr/bin/git'
def gitexec(command, *args):
    stat, output, err = pexecl(GITPATH, 'git', command, *args)
    if stat: raise VCSException, ('git %s failed'%command, args, stat, output, err)
    return output

class Git(VCS):
    @classmethod
    def _create(self):
        self._new = not os.path.exists('.git')
        gitexec('init')

    def _current_revision(self, file):
        output = gitexec('log', '-1', '--pretty=format:%H')
        return output.strip()

    # see GIT-EXTERNAL-DIFF in git(7)
    def _commit_message(self, id):
        output = gitexec('log', '-1', id)
        lines = output.split('\n')
        blank = lines.index('')

        lines = lines[blank+1:-1]
        lines = [line[4:].rstrip() for line in lines]

        message = '\n'.join(lines)

        return message

    def _commit_timestamp(self, id):
        output = gitexec('log', '-1', id, '--pretty=format:%ct')
        return output.strip()

    def _unparse_timestamp(self, ts):
        return datetime.datetime.fromtimestamp(int(ts))

    def _get(self, filename, revid):
        if not revid:
            try:
                return file(filename).read()
            except IOError, e:
                # Crappy git behavior
                if e.errno == errno.EISDIR: return ''
                # Falling through to git

        if revid == None: revid = 'HEAD'
        return gitexec('show', revid+':'+filename)

    def _stream(self, filename, revid):
        # FIXME: Some magic way to read git's packs or object DB?
        if revid == None: revid = 'HEAD'
        f = pexecl_stream(GITPATH, 'git', 'show', revid+':'+filename)
        return f

    def _history(self, filenames, rev_from, rev_to, max):
        if rev_from == None: rev_from = ''
        if rev_to == None: rev_to = ''
        args = ['--pretty=format:%H']
        if rev_from or rev_to:
            args += [rev_from + '..' + rev_to]
        if max:
            args += ['-'+str(max)]
        args += filenames
        output = gitexec('log', *args)
        revisions = output.split()
        return [rev.strip() for rev in revisions]

    def _listdir(self, directory, revid):
        # FIXME: a lot of this might become faster if we switched to Dulwich
        if revid == None:
            if directory == '': directory = '.'
            try:
                lst = os.listdir(directory)
            except OSError, e:
                if e.errno == errno.ENOENT:
                    raise VCSException, "directory doesn't exist"
                raise e
            # FIXME: git won't include empty directories here!
            # Strip them out manually?
            # Ensure they don't happen?
            # Handle them at the query level?
            for f in lst[:]:
                if f.startswith('.'): lst.remove(f)
            return lst

        if revid == None: revid = "HEAD"
        directory = revid + ':' + directory
        output = gitexec('ls-tree', directory)
        lines = output.split('\n')
        lines.pop() # strip trailing newline
        files = []
        for line in lines:
            fields = line.split()
            files.append(fields[-1])
        return files

    def _isdir(self, filename, revid):
        if revid == None: revid = "HEAD"
        return gitexec('cat-file', '-t', revid+':'+filename) == 'tree'

    def _add(self, files):
        gitexec('add', *files)

    def _commit(self, files, message):
        if files:
            # this shouldn't happen; git-commit filename automatically
            # does a git-add if the file is known to git already. If
            # git doesn't already know the file, why are you
            # committing it without adding it?
            self._add(files)
        gitexec('commit', '-m', message, *files)

    def _move(self, src, dest):
        args = ['mv']
        args.extend(src)
        args.append(dest)
        return gitexec(*args)

    def _diff(self, src, dest, fname=None):
        return gitexec('diff', '--color-words', src+'..'+dest, fname or '')

    def _remove(self, src):
        return gitexec('rm', '-r', src)
