from __future__ import with_statement
from lockers import application
import lockers.node
import lockers.diff as diff
from lockers.format.rst import RST
from webob import Response
import webob.exc
import os
from lockers.vcs import FileNotFoundError, VCSException
from lockers.models.models import Changeset, LockersObject, DoesNotExistError, ModelFactory, ChangesetFactory
from lockers.models.fields import StrPart
from lockers.template import render_to_response
import traceback

class WikiChangeset_1(Changeset):
    pass

WikiChangeset = ChangesetFactory('WikiChangeset')

class WikiObject_1(LockersObject):
    '''A wiki object has the following parts:

    index: contents, as YAML'd RST
    '''
    COMMENT = WikiChangeset
    index = StrPart()

WikiObject = WikiObject_1

class Wiki(application.LockersApplication):
    '''Wiki module for lockers.'''
    def __init__(self):
        super(Wiki, self).__init__()
        self.rst = RST()

    def __call__(self, request):
        super(Wiki, self).__call__(request)
        node = request.node
        #print request.application_url

        if ('action', 'edit') in request.GET.items():
            return self.edit(request, node)
        elif ('action', 'history') in request.GET.items():
            return self.history(request, node)
        elif ('action', 'diff') in request.GET.items():
            return self.diff(request, node)

        else: return self.view(request, node)

    def edit(self, request, node):
        if request.method == 'POST':
            # save a new version
            with self.backend.new_transaction() as t:
                obj = WikiObject.get(t, node)
                obj.index = request.POST.getone('content')
                obj.save({'message': request.POST.getone('message')})

            url = node.url_to()

            return webob.exc.HTTPFound(detail='Post saved.', location=url)

        return self.render_to_response(request, 'wiki/edit.html', {
                'page_name': node.path,
                'page_text': self.get(node),
                })

    def get(self, node):
        try:
            with self.backend.new_transaction() as t:
                obj = WikiObject(t, node)
                filetext = obj.index
        except (VCSException, DoesNotExistError), e:
            #print e, id(e), e.message
            traceback.print_exc(e)
            filetext = 'new wikipage'
        return filetext

    def format(self, path):
        return self.rst.format(self.get(path))

    def view(self, request, node):
        text = self.format(node)

        return self.render_to_response(request, 'wiki/view.html', {
                'page_name': node.path,
                'page_text': text,
                'node': node,
                })

    def history(self, request, node):
        with self.backend.new_transaction() as t:
            revids = t.history(node)
            
        versions = []
        for rev in revids:
            versions.append({'id': rev, 'link': node.url_to(revision=rev)})

        return self.render_to_response(request, 'wiki/versions.html', {
                'node': node,
                'versions': versions
                })

    def diff(self, request, node):
        post = request.GET
        revid1 = post.getone('diff1')
        revid2 = post.getone('diff2')

        with self.backend.new_transaction() as t:
            p1 = WikiObject(t, node.replace(revision=revid1)).index
            p2 = WikiObject(t, node.replace(revision=revid2)).index

            diff = t.diff(revid1, revid2, node)

        return self.render_to_response(request, 'wiki/diff.html', {
                'node': node,
                'version1': revid1, 'version2': revid2,
                'diff': diff.FormatDiff(p1, p2).render()
                })

wiki = Wiki
