from __future__ import with_statement
import os.path
import syck

import TestSupport
import TestApplication

from lockers.models.models import LockersObject, UntrackedObject, ModelFactory, Changeset, ChangesetFactory
from lockers.models.utils import tuple_to_list
import lockers.backend as backend
from lockers.node import Node
import lockers.vcs as vcs
import lockers.models.fields as fields

class ModelsObject_1(LockersObject):
    foo = fields.StrField()

class ModelsObject_2(ModelsObject_1):
    bar = fields.NumField()

class ModelsObject_3(ModelsObject_2):
    baz = [fields.NumField()]

ModelsObject = ModelFactory('ModelsObject')

class SubClassObject_1(ModelsObject):
    quux = fields.NumField()

class SubClassObject_2(SubClassObject_1):
    quuux = fields.StrField()

SubClassObject = ModelFactory('SubClassObject')

class ModelsUntrackedObject_1(UntrackedObject):
    bar = fields.StrField()

ModelsUntrackedObject = ModelFactory('ModelsUntrackedObject')

class ExpandObject_1(LockersObject):
    add_extra_fields = True

ExpandObject = ModelFactory('ExpandObject')

class UntrackedEObject_1(UntrackedObject):
    add_extra_fields = True

UntrackedEObject = ModelFactory('UntrackedEObject')

class FileTest_1(LockersObject):
    class meta(fields.Part):
        filename = fields.StrField()

class FileTest_2(FileTest_1):
    class meta(FileTest_1.meta):
        sha1 = fields.StrField()

class ModelsTestChangeset_1(Changeset):
    old_name = fields.StrField()
    new_name = fields.StrField()

class ModelsTestChangeset_2(ModelsTestChangeset_1):
    timestamp = fields.StrField()

ModelsTestChangeset = ChangesetFactory('ModelsTestChangeset')

class DiffChangeset_1(Changeset):
    diff_field = fields.DiffField()

DiffChangeset = ChangesetFactory('DiffChangeset')

class ModelsTestSubChangeset_1(ModelsTestChangeset):
    sigil = fields.NumField()

ModelsTestSubChangeset = ChangesetFactory('ModelsTestSubChangeset')

def get_modelsclassver(fromno):
    '''Utility to get the nth numbered ModelsObject class version'''
    return (fromno, LockersObject.classes.latest)

def get_subclassver(from_modno, from_subno):
    '''Utility to get the (m, n)th numbered SubClassObject class version'''
    return (from_subno, get_modelsclassver(from_modno))

class ModelsTests(TestSupport.TestCase):
    def setUp(self):
        TestApplication.init('/tmp/backendtest')
        self.backend = backend.get_backend()

    def test_factory(self):
        self.assertEqual(len(ModelsObject._meta.multiclass), 3)

    def fields_equal(self, cls, fieldnames):
        fields = cls._meta.get_field('index')._meta.fields
        #print 'checking for equality of', fields, fieldnames
        seen = {}
        for f in fields:
            if f.name not in fieldnames:
                raise AssertionError, '%s not in %s' % (f.name, fieldnames)
            seen[f.name] = True

        for f in fieldnames:
            if f not in seen:
                raise AssertionError, "didn't see %s in %s" % (f, fields)
                return False

        return True

    def test_inherit_factory(self):
        m = get_modelsclassver
        s = get_subclassver
        self.assertEqual(ModelsObject.classes.get(m(1)), ModelsObject_1)
        self.assertEqual(ModelsObject.classes.get(m(2)), ModelsObject_2)
        self.assertEqual(ModelsObject.classes.get(m(3)), ModelsObject_3)

        self.fields_equal(SubClassObject.classes.get(s(1, 1)),
                          ['quux', 'foo'])

        self.fields_equal(SubClassObject.classes.get(s(2, 1)),
                          ['quux', 'foo', 'bar'])

        self.fields_equal(SubClassObject.classes.get(s(3, 1)),
                          ['quux', 'foo', 'bar', 'baz'])

        self.fields_equal(SubClassObject.classes.get(s(1, 2)),
                          ['quux', 'quuux', 'foo'])

        self.fields_equal(SubClassObject.classes.get(s(2, 2)),
                          ['quux', 'quuux', 'foo', 'bar'])

        self.fields_equal(SubClassObject.classes.get(s(3, 2)),
                          ['quux', 'quuux', 'foo', 'bar', 'baz'])

        self.fields_equal(ModelsObject.classes.get(m(1)),
                          ['foo'])
        self.fields_equal(ModelsObject.classes.get(m(2)),
                          ['foo', 'bar'])
        self.fields_equal(ModelsObject.classes.get(m(3)),
                          ['foo', 'bar', 'baz'])

        # FIXME: these may need to get changed if we ever expand the example
        self.assertEqual(len(ModelsObject.classes.versions), 3*len(LockersObject.classes.versions))
        self.assertEqual(len(SubClassObject.classes.versions), 6*len(LockersObject.classes.versions))

        node = Node('objecttest', '/subclass/file1')
        with self.backend.new_transaction() as t:
            o = SubClassObject.new(t, node)
            o.save(Changeset({'message': 'Hi'}))

        # SubClassObject version: (2, (3, (1,))).
        # i.e. SubClassObject version 2 based on ModelsObject 3 based
        # on LockersObject version 1.
        # - name: SubClassObject
        #   version: 2
        # - - name: ModelsObject
        #     version: 3
        #   - - name: LockersObject
        #       version: 1
        l = syck.load(file(node.part('lockersmeta')))['version']
        self.assertEqual(l[0], {'version': 2, 'name': 'SubClassObject'})
        self.assertEqual(l[1][0], {'version': 3, 'name': 'ModelsObject'})
        self.assertEqual(l[1][1][0], {'version': 1, 'name': 'LockersObject'})

        s = file(node.part('lockersmeta')).read()
        if '!python/tuple' in s:
            print 'Complaint: ugly tuple stuff in the git repo'

    def make_invalid_class1(self):
        class MyFunnyObject(LockersObject):
            pass

    def make_invalid_class2(self):
        class MyFunnyObject_foo(LockersObject):
            pass

    def test_invalid_class(self):
        self.assertRaises(TypeError, self.make_invalid_class1)
        self.assertRaises(TypeError, self.make_invalid_class2)


    def test_object(self):
        node = Node('objecttest', '/file1')
        node2 = Node('objecttest', '/file2')
        with self.backend.new_transaction() as t:
            self.assertRaises(backend.DoesNotExistError, lambda: LockersObject(t, node))
            o = ExpandObject.new(t, node)
            o.part1 = u'This is some text'
            o.save(Changeset({'message': 'Commit message'}))

            o2 = ExpandObject.get(t, node2)
            o2.part1 = 'I was new here'
            o2.save(Changeset({'message': 'Create a new object using get'}))

        with self.backend.new_transaction() as t:
            self.assertRaises(backend.AlreadyExistsError, lambda: LockersObject.new(t, node))
            o12 = ExpandObject(t, node)
            self.assertEqual(o12.part1, 'This is some text')

            o22 = ExpandObject.get(t, node2)
            self.assertEqual(o22.part1, 'I was new here')

    def test_untracked(self):
        node = Node('objecttest', '/dir/untracked1', untracked=True)
        with self.backend.new_transaction() as t:
            rev = t.current_revision()
            self.failIf(t.exists(node))
            self.assertRaises(vcs.VCSException, lambda: t.history([node]))
            u = UntrackedEObject.new(t, node)
            u.part1 = 'some text'
            u.save()

        with self.backend.new_transaction() as t:
            rev2 = t.current_revision()
            u_load = UntrackedEObject(t, node)
            self.assertEqual(u_load.part1, 'some text')
            self.assertEqual(t.history([node]), [])
            self.assert_(t.exists(node))
            self.assertEqual(rev, rev2)
            self.assertRaises(TypeError, lambda: UntrackedObject(t, node))


    def test_constructor(self):
        import pprint
        node = Node('objecttest', '/file2')
        self.assert_(not hasattr(ModelsObject_2, 'baz'))
        self.assert_(FileTest_2.meta)

        with self.backend.new_transaction() as t:
            m = ModelsObject_2.new(t, node, bar=1)
            self.assert_(hasattr(m.index, 'bar'))
            self.assert_(hasattr(m.index, 'foo'))
            self.assertEqual(m.index.bar, 1)

            m3 = ModelsObject_3.new(t, node, baz=[1, 2])
            self.assertEqual(m3.index.baz, [1, 2])

            m3_1 = ModelsObject_3.new(t, node)
            self.assertEqual(m3_1.index.baz, [])

            f = FileTest_2.new(t, node,
                           meta=FileTest_2.meta(filename="hi", sha1="abcdabcd"))
            self.assertEqual(f.meta.filename, 'hi')
            self.assertEqual(f.meta.sha1, 'abcdabcd')


    def test_validation(self):
        e = fields.EmailField()
        e.validate('example@example.com', []) # should raise on failure
        self.assertRaises(fields.ValidationError, lambda:
                          e.validate('foobar@', []))

    def valid_models1(self, o):
        self.assertEqual(o.index.foo, '')
        self.assert_(not hasattr(o.index, 'bar'))
        self.assert_(not hasattr(o.index, 'baz'))
        self.assert_(isinstance(o, ModelsObject_1))

    def valid_models2(self, o):
        self.assert_(hasattr(o.index, 'bar'))
        self.assert_(not hasattr(o.index, 'baz'))
        self.assert_(isinstance(o, ModelsObject_2))

    def valid_models3(self, o):
        self.assert_(hasattr(o.index, 'bar'))
        self.assert_(hasattr(o.index, 'baz'))
        self.assert_(isinstance(o, ModelsObject_3))

    def test_migration(self):
        node = Node('objecttest', '/models/objects/version1')
        with self.backend.new_transaction() as t:
            o1 = ModelsObject_1.new(t, node)
            self.valid_models1(o1)

            o2 = ModelsObject.classes.migrate(o1, to=(2, 1))
            self.valid_models2(o2)
            self.valid_models2(o1)

            o3 = ModelsObject.classes.migrate(o2, to=(3, 1))
            self.valid_models3(o3)
            self.valid_models3(o2)
            self.valid_models3(o1)


    def test_loading(self):
        node = Node('objecttest', '/models/objects/start1')
        with self.backend.new_transaction() as t:
            o1 = ModelsObject_1.new(t, node)
            self.valid_models1(o1)
            o1.index.foo = 'foo on you'
            o1.save(Changeset({'message': 'hi there'}))

        with self.backend.new_transaction() as t:
            o = ModelsObject.get(t, node)
            self.assert_(isinstance(o, ModelsObject_1))

            self.assertRaises(TypeError, ModelsObject_2.get, t, node)
            # FIXME: decided this shouldn't be allowed
#             o2 = ModelsObject_2.get(t, node)
#             self.assert_(isinstance(o2, ModelsObject_2))

#             o2.save(Changeset({'message': 'yo'}))

#         with self.backend.new_transaction() as t:
#             o = ModelsObject.get(t, node)
#             self.assert_(isinstance(o, ModelsObject_2))

            self.assertRaises(TypeError, LockersObject.get, t, node)


    def test_exception_without_version(self):
        def makeNewClass():
            # should raise exception because it doesn't have a version
            class Foo(LockersObject):
                pass

        self.assertRaises(TypeError, makeNewClass)

    def test_properties(self):
        node = Node('objecttest', '/models/objects/index1')
        with self.backend.new_transaction() as t:
            o = ModelsObject_3.new(t, node)
            o.foo = 'hey'
            o.bar = 41
            o.baz = [18]
            self.assertEqual(o.foo, o.index.foo)
            self.assertEqual(o.bar, o.index.bar)
            self.assertEqual(o.baz, o.index.baz)
            o.baz.append(19)
            self.assertEqual(o.baz, [18, 19])
            o.save({'message': "testing"})

        with self.backend.new_transaction() as t:
            o2 = ModelsObject_3(t, node)
            for attr in ['foo', 'bar', 'baz']:
                self.assertEqual(getattr(o, attr), getattr(o2, attr))
                self.assertEqual(getattr(o, attr), getattr(o2.index, attr))
                self.assertEqual(getattr(o.index, attr), getattr(o2, attr))
                self.assertEqual(getattr(o.index, attr), getattr(o2.index, attr))


            self.assertRaises(TypeError, lambda: setattr(o2, 'quux', 24))

    def test_invalid(self):
        # Make sure certain invalid classes get exceptions
        def def1():
            class Foo_1(LockersObject):
                lockersmeta = fields.Part()

        self.assertRaises(TypeError, def1)

    def test_delete(self):
        n1 = Node('objecttest', '/models/objects/deleteable')
        n2 = Node('objecttest', '/models/objects/deleteable-2', untracked=True)
        p1 = 'objecttest/models/objects/deleteable'
        p2 = 'objecttest/models/objects/deleteable-2'

        with self.backend.new_transaction() as t:
            revid1 = t.current_revision()
            o = ModelsObject_3.new(t, n1)
            o.foo = 'hey'
            o.bar = 41
            o.baz = [18]

            o.save(Changeset({'message': 'object created'}))

        with self.backend.new_transaction() as t:
            revid2 = t.current_revision()
            self.failIf(revid1 == revid2)
            self.assert_(t.exists(n1))
            self.assert_(os.path.exists(p1))

            # FIXME: this API sucks.
            # We don't really want to load the object before we delete it, do we?
            t.remove(n1)
            t.commit(Changeset({'message': 'object deleted'}).as_commit_message())

        with self.backend.new_transaction() as t:
            revid3 = t.current_revision()
            self.failIf(revid2 == revid3)
            self.failIf(t.exists(n1))

        with self.backend.new_transaction() as t:
            revid1 = t.current_revision()
            o = ModelsUntrackedObject.new(t, n2)
            o.bar = 'abcd'
            o.save()

        with self.backend.new_transaction() as t:
            revid2 = t.current_revision()
            self.assert_(t.exists(n2))
            self.assertEqual(revid1, revid2)
            self.assert_(os.path.exists(p2))
            t.remove(n2)

        with self.backend.new_transaction() as t:
            revid3 = t.current_revision()
            self.assertEqual(revid2, revid3)
            self.failIf(t.exists(n2))

        # FIXME: This doesn't work. After all, there could be children of n1/n2
        #self.failIf(os.path.exists(p1))
        #self.failIf(os.path.exists(p2))

    def test_changesets(self):
        node = Node('changesettest', '/ethan/some/object')
        with self.backend.new_transaction() as t:
            rev1 = t.current_revision()
            o = ModelsObject.new(t, node)
            d1 = {'message': None, 'old_name': 'foo', 'new_name': 'bar'}
            o.save(ModelsTestChangeset_1(d1))

        with self.backend.new_transaction() as t:
            rev2 = t.current_revision()
            o = ModelsObject.get(t, node)
            o.foo = 'changing foo'
            d2 = {'message': 'Just for fun', 'old_name': 'bar',
                  'new_name': 'baz', 'timestamp': 'right now'}
            o.save(ModelsTestChangeset_2(d2))

        with self.backend.new_transaction() as t:
            rev3 = t.current_revision()
            o = ModelsObject.get(t, node)
            o.foo = 'newest foo'
            d3 = {'message': 'Another commit', 'old_name': 'baz',
                  'new_name': 'fred', 'timestamp': 'yesterday',
                  'sigil': 1}
            o.save(ModelsTestSubChangeset(d3))

        rev4 = self.backend.current_revision()

        with self.backend.new_transaction() as t:
            #c1 = Changeset.load(t, rev1)
            #assert not isinstance(c1, ModelsTestChangeset_1)  # who knows what it is

            def compare(changeset, dict):
                for key in dict:
                    self.assertEqual(getattr(changeset, key), dict[key],
                                     'attribute %s is wrong in %s: should be %s, got %s'%(key, changeset, dict[key], getattr(changeset, key)))

            c2 = Changeset.load(t, rev2)
            assert isinstance(c2, ModelsTestChangeset_1)
            #assert not isinstance(c2, ModelsTestChangeset_2) # FAILS: "migrated" to newest
            assert ModelsTestChangeset.isinstance(c2)
            assert not isinstance(c2, ModelsTestSubChangeset_1)
            assert not ModelsTestSubChangeset.isinstance(c2)
            compare(c2, d1)

            c3 = Changeset.load(t, rev3)
            assert isinstance(c3, ModelsTestChangeset_2)
            assert ModelsTestChangeset.isinstance(c3)
            assert not isinstance(c3, ModelsTestSubChangeset_1)
            assert not ModelsTestSubChangeset.isinstance(c3)
            compare(c3, d2)

            c3.old_name = 'moo'
            c3_2 = Changeset.load(t, rev3)
            compare(c3_2, d2)
            # Ensure that c3 hasn't changed with c3_2 being loaded.
            # I actually had a bug like that, except even stupider!
            self.assertEqual(c3.old_name, 'moo')

            c4 = Changeset.load(t, rev4)
            assert isinstance(c4, ModelsTestSubChangeset_1)
            assert ModelsTestChangeset.isinstance(c4)
            assert ModelsTestSubChangeset.isinstance(c4)
            compare(c4, d3)

            compare(c3, d2)

    def test_dependencies(self):
        node1 = Node('module1', 'object1')
        cnode1 = Node('cached', 'module1/object1/form1', untracked=True)
        with self.backend.new_transaction() as t:
            cobj = ModelsUntrackedObject.new(t, cnode1)
            cobj.bar = 'an object'
            cobj.save()

            obj = ModelsObject_1.new(t, node1)
            obj.index.foo = 'some text\n'
            obj.save(Changeset({'message': 'creating node1'}))

        with self.backend.new_transaction() as t:
            assert t.exists(cnode1)
            cobj = ModelsUntrackedObject.existing(t, cnode1)
            cobj.register_dependencies([node1])

        with self.backend.new_transaction() as t:
            assert t.exists(cnode1)
            obj = ModelsObject_1.existing(t, node1)
            obj.index.foo = 'now text'
            obj.save(Changeset({'message': 'changing node'}))

        with self.backend.new_transaction() as t:
            assert not t.exists(cnode1)


    def test_diff_changeset(self):
        # I would like to be able to generate these diffs too, but for now this
        # is fine
        import syck
        c = DiffChangeset({'diff_field': {'old': 'hi', 'new': 'bye'}})
        self.assertEqual(c.diff_field.old, 'hi')
        self.assertEqual(c.diff_field.new, 'bye')
        flat = syck.load(c.serialize())
        self.assertEqual(flat['diff_field'], {'old': 'hi', 'new': 'bye'})
        c2 = DiffChangeset(flat)
        self.assertEqual(c.diff_field, c2.diff_field)
        self.assertEqual(c.serialize(), c2.serialize())

        c_n = DiffChangeset({'diff_field': None})
        self.assertEqual(c_n.diff_field, None)
        flat = syck.load(c_n.serialize())
        self.assertEqual(flat['diff_field'], None)
        c_n2 = DiffChangeset(flat)
        self.assertEqual(c_n.diff_field, c_n2.diff_field)
        self.assertEqual(c_n.serialize(), c_n2.serialize())


TestSupport.run(__name__)
