import TestSupport
import lockers.node as node
from lockers.request import LockersRequest

class NodeTests(TestSupport.TestCase):
    def test_url(self):
        n = node.Node('wiki', '/')
        self.assertEqual(n.url().href, '/wiki/')

        n2 = node.Node('wiki', '/some/path')
        self.assertEqual(n2.url().href, '/wiki/some/path')

    def test_replace(self):
        n = node.Node('wiki', '/Memes/Series_of_Tubes')
        n2 = n.replace(path='/Series/Baccano!')
        # Make sure the args are erased
        self.assertEqual(n2.url().href, '/wiki/Series/Baccano!')

        n2 = n.replace(module='games')
        # Both path and args should be erased
        self.assertEqual(n2.url().href, '/games/')

        self.assertRaises(ValueError, lambda: n.replace(args={}))
        self.assertRaises(ValueError, lambda: n.replace(action='edit'))

    def test_path_safety(self):
        n = node.Node('wiki', 'hi')
        n2 = node.Node('wiki', '/hi')
        self.assertEqual(n, n2)

    def test_filename(self):
        n = node.Node('wiki', '/Memes/')
        n2 = node.Node('wiki', '/Memes')
        n3 = node.Node('wiki', 'Memes')
        self.assertEqual(n.filename(), 'wiki/Memes/index')
        self.assertEqual(n2.filename(), n.filename())
        self.assertEqual(n3.filename(), n.filename())

        self.assertEqual(n.part('index'), 'wiki/Memes/index/index')
        self.assertEqual(n2.part('index'), n.part('index'))
        self.assertEqual(n3.part('index'), n.part('index'))


    WIKI_TUBES = '/wiki/Memes/Series_of_Tubes'
    def request_node(self, path):
        request = LockersRequest.blank(path)
        request.path_info_pop()  # set module correctly
        n = node.Node(request)
        return n

    def test_request(self):
        tubes = self.WIKI_TUBES
        n = self.request_node(tubes)
        self.assertEqual(n.url().href, tubes)

        tubes_edit = tubes + '?action=edit'

        n = self.request_node(tubes_edit)
        # node reflects a "file" or object; almost every query argument is a
        # parameter to a controller (rather than the model)
        self.assertEqual(n.url().href, tubes)

        # But at the same time, a node can be used to generate URLs. Should we
        # maybe support this differently?
        u = n.url_to(action='edit')
        self.assertEqual(u, tubes_edit)

    def test_url_to(self):
        tubes = self.WIKI_TUBES
        n = self.request_node(tubes)
        n2 = node.Node(module='files', path='/some/file/id')

        u = n.url_to(n2, action='edit')
        self.assertEqual(u, '/files/some/file/id?action=edit')

        n2 = node.Node(path='/Memes/Yo_Dawg')
        u = n.url_to(n2, action='edit')
        self.assertEqual(u, '/wiki/Memes/Yo_Dawg?action=edit')

        self.assertRaises(TypeError, lambda: n.url_to(n2, n2))
        self.assertEqual(n.url_to(n2), n.url_to(path=n2.path))
        u = n.url_to(n2, action='edit', path='/Memes/You_Insensitive_Clod')
        self.assertEqual(u, '/wiki/Memes/You_Insensitive_Clod?action=edit')

        n.revision = 'foo'
        u = n.url_to(n2)
        self.assertEqual(u, '/wiki/Memes/Yo_Dawg?revision=foo')

    def test_push(self):
        n = node.Node('wiki', 'Memes')
        norig = node.Node('wiki', 'Memes')
        npushed = n.push('tags')
        nres = node.Node('tags', 'wiki/Memes')
        self.assertEqual(npushed.filename(), nres.filename())
        self.assertEqual(npushed.part('index'), nres.part('index'))
        self.assertEqual(n.filename(), norig.filename())
        self.assertEqual(n.part('index'), norig.part('index'))

TestSupport.run(__name__)
