import os
import TestSupport
import lockers.vcs as vcs
import lockers.utils as utils

def dump(filename, contents):
    file(filename, 'w').write(contents)

class VCSTests(TestSupport.TestCase):
    TESTREPO = 'vcsrepo'
    def setUp(self):
        os.mkdir(self.TESTREPO)
        os.chdir(self.TESTREPO)
        self.vcs = vcs.Git.create()

    def tearDown(self):
        os.chdir('..')
        os.system('rm -rf '+self.TESTREPO)

    def test_commit(self):
        '''test that a commit stores a file and a message'''
        fname = 'file1'
        contents = 'initial contents\n'
        message = 'initial message'
        dump(fname, contents)
        self.vcs.add(fname)
        self.vcs.commit(fname, message)
        vcscontents = self.vcs.get(fname, self.vcs.current_revision())
        self.assertEqual(vcscontents, contents) # contents changed when stored?
        vcsmessage = self.vcs.commit_message(self.vcs.current_revision())
        self.assertEqual(vcsmessage, message) # message changed when stored?


    def test_revisions(self):
        '''test that history of a file selects the right revisions, and not the wrong ones'''
        fname1 = 'file1'
        fname2 = 'file2'
        dump(fname1, '1\n')
        self.vcs.add(fname1)
        self.vcs.commit(fname1, 'rev1')
        revid1 = [self.vcs.current_revision()]
        dump(fname2, '1\n')
        self.vcs.add(fname2)
        self.vcs.commit(fname2, 'rev2')
        revid2 = [self.vcs.current_revision()]
        # Does current revision change?
        self.assert_(revid1[0] != self.vcs.current_revision())
        # Is the current revision of file2 horribly wrong?
        self.assert_(revid1[0] != self.vcs.current_revision(fname2))
        # Does current revision of a file return correct revision?
        self.assertEqual(self.vcs.current_revision(), self.vcs.current_revision(fname2))
        # Is the current revision of file2 correct?
        self.assertEqual(revid2[0], self.vcs.current_revision(fname2))

        dump(fname1, '2\n')
        self.vcs.commit(fname1, 'rev3')
        revid1.append(self.vcs.current_revision())

        vcsrevs1 = self.vcs.history([fname1])
        vcsrevs1.reverse() # vcs returns history in reverse-chronological
        # Is the history of file1 correct?
        self.assertEqual(vcsrevs1, revid1)

        self.assertEqual(self.vcs.listdir(), [fname1, fname2])


    def test_get(self):
        '''Test that when we get previous revisions, horrible things don't happen.'''
        fname1 = 'file1'
        fname1_new = 'file1-new'
        dump(fname1, '1\n')
        self.vcs.add(fname1)
        self.vcs.commit(fname1, 'rev1')
        revid1 = self.vcs.current_revision()

        self.vcs.move(fname1, fname1_new)
        # We can't use the filenames here because file1 isn't in this revision
        # Should all commits be [] like this?
        self.vcs.commit([], 'rev2')          # FIXME: this sucks
        revid2 = self.vcs.current_revision()

        self.assert_(not self.vcs.exists(fname1))
        self.assert_(not self.vcs.exists(fname1, revid2))
        self.assert_(self.vcs.exists(fname1, revid1))

        self.assert_(self.vcs.exists(fname1_new))
        self.assert_(self.vcs.exists(fname1_new, revid2))
        self.assert_(not self.vcs.exists(fname1_new, revid1))

    def test_delete(self):
        fname1 = 'my-deleteable'
        dump(fname1, '1\n')
        self.vcs.add(fname1)
        self.vcs.commit(fname1, 'rev1')
        revid1 = self.vcs.current_revision()

        self.assert_(self.vcs.exists(fname1))
        self.assert_(self.vcs.exists(fname1, revid1))

        self.vcs.remove(fname1)
        self.vcs.commit([], 'rev2') # FIXME: this sucks too
        revid2 = self.vcs.current_revision()

        self.failIf(self.vcs.exists(fname1))
        self.failIf(self.vcs.exists(fname1, revid2))

        self.assert_(self.vcs.exists(fname1, revid1))



TestSupport.run(__name__)

